"""metaknowledge is a Python3 package that simplifies bibliometric and computational analysis of Web of Science data.

# Example

To load the data from files and make a network:

    >>> import metaknowledge as mk
    >>> RC = mk.RecordCollection("records/")
    >>> print(RC)
    Collection of 33 records
    >>> G = RC.coCiteNetwork(nodeType = 'journal')
    Done making a co-citation network of files-from-records                 1.1s
    >>> print(len(G.nodes()))
    223
    >>> mk.write_graph(G, "Cocitation-Network-of-Journals")

# Overview

This package can read the files downloaded from the [Thomson Reuters Web of Science](https://webofknowledge.com) (WOS) as plain text. These files contain metadata about scientific records, such as the authors, language, and citations. The records are saved in groups of up-to 500 individual records in a file.

The [metaknowledge.RecordCollection](#RecordCollection.RecordCollection) class can take a path to one or more of these files load and parse them. The object is the main way for work to be done on multiple records. For each individual record it creates an instance of the [metaknowledge.Record](#Record.Record) class that contains the results of the parsing of the record.

The files given by WOS are a flat database containing a series of 2 character tags, e.g. 'TI' is the title. Each WOS tag has one or more values and metaknowledge makes use of them to extract useful information. The approximate meanings of the tags are listed in the [tagFuncs](#tagFuncs.tagFuncs) package, if you simply want the mapping `tagToFull` is a dictionary that will provide it. There are no full official public listings of tag the meanings available. metaknowledge is not attempting to provide the definitive or authoritative meanings.

As citations are of great importance to sociology their handling is done with the [Citation](#Citation.Citation) class. This class can parse the citations given by WOS as well as extra details about the full name of their journal and allow simple comparisons.

Note for those reading the docstring metaknowledge's docs are written in markdown and are processed to produce the documentation found at [networkslab.org/metaknowledge/documentation](http://networkslab.org/metaknowledge/documentation/).
"""
from .record import Record, recordParser, BadISIRecord
from .citation import Citation, BadCitation, filterNonJournals
from .recordCollection import RecordCollection, isiParser
from .graphHelpers import write_edgeList, write_nodeAttributeFile, write_graph, read_graph, _ProgressBar, drop_edges, drop_nodesByDegree, drop_nodesByCount, graphStats
from .constants import VERBOSE_MODE, tagToFull, fullToTag, tagNameConverter, tagsAndNames, knownTagsList
#from .blondel import blondel, modularity
from .diffusion import diffusionGraph, diffusionCount



import networkx as nx
import matplotlib.pyplot as plt
import numpy as np
import scipy.ndimage as ndi
import math

def quickVisual(G, showLabel = False):
    """just makes a simple matplolib figure and displays it, with each node coloured by its type"""
    colours = "brcmykwg"
    f = plt.figure(1)
    ax = f.add_subplot(1,1,1)
    ndTypes = []
    ndColours = []
    layout = nx.spring_layout(G, k = 4 / math.sqrt(len(G.nodes())))
    for nd in G.nodes(data = True):
        if 'type' in nd[1]:
            if nd[1]['type'] not in ndTypes:
                ndTypes.append(nd[1]['type'])
            ndColours.append(colours[ndTypes.index(nd[1]['type']) % len(colours)])
        elif len(ndColours) > 1:
            raise RuntimeError("Some nodes do not have a type")
    if len(ndColours) < 1:
        nx.draw_networkx_nodes(G, pos = layout, node_color = colours[0], node_shape = '8', node_size = 100, ax = ax)
    else:
        nx.draw_networkx_nodes(G, pos = layout, node_color = ndColours, node_shape = '8', node_size = 100, ax = ax)
    nx.draw_networkx_edges(G, pos = layout, width = .7, ax = ax)
    if showLabel:
        nx.draw_networkx_labels(G, pos = layout, font_size = 8, ax = ax)
    plt.axis('off')
    f.set_facecolor('w')

def graphDensityContourPlot(G, layout = None, layoutScaleFactor = 1, shifAxis = False, overlay = False, axisSamples = 100, blurringFactor = .1, contours = 15, nodeSize = 10, graphType = 'coloured', iters = 50):
    """
    Requires numpy and matplotlib

    graphType is either "coloured or "solid"
    """
    from mpl_toolkits.mplot3d import Axes3D

    if not isinstance(G, nx.classes.digraph.DiGraph) and not isinstance(G, nx.classes.graph.Graph):
        raise TypeError("{} is not a valid input.".format(type(G)))
    if layout is None:
        layout = nx.spring_layout(G, scale = axisSamples - 1, iterations = iters)
        grid = np.zeros( [axisSamples, axisSamples],dtype=np.float32)
        for v in layout.values():
            x, y = tuple(int(x) for x in v.round(0))
            grid[y][x] += 1
    elif isinstance(layout, dict):
        layout = layout.copy()
        grid = np.zeros([axisSamples, axisSamples],dtype=np.float32)
        multFactor = (axisSamples - 1) / layoutScaleFactor
        for k in layout.keys():
            if shifAxis:
                tmpPos = (layout[k]  + layoutScaleFactor / 2) * multFactor
            else:
                tmpPos = layout[k] * multFactor
            layout[k] = tmpPos
            x, y = tuple(int(x) for x in tmpPos.round(0))
            grid[y][x] += 1
    else:
        raise TypeError("{} is not a valid input.".format(type(layout)))

    fig = plt.figure()
    #axis = fig.add_subplot(111)
    axis = fig.gca(projection='3d')
    if overlay:
        nx.draw_networkx(G, pos = layout, ax = axis, node_size = nodeSize, with_labels = False, edgelist = [])
    grid = ndi.gaussian_filter(grid, (blurringFactor * axisSamples, blurringFactor * axisSamples))
    X = Y = np.arange(0, axisSamples, 1)
    X, Y = np.meshgrid(X, Y)
    if graphType == "solid":
        CS = axis.plot_surface(X,Y, grid)
    else:
        CS = axis.contourf(X, Y, grid, contours)
    axis.set_xlabel('X')
    axis.set_ylabel('Y')
    axis.set_zlabel('Node Density')



import metaknowledge

import networkx as nx
import csv
import os
import sys
import time
import math
import threading

def read_graph(edgeList, nodeList = None, directed = False, idKey = 'ID', eSource = 'From', eDest = 'To'):
    """Reads the files given by edgeList and if given nodeList. Outputs a networkx graph for the lists.

    This is designed only for the files produced by metaknowledge and is meant to be the reverse of [write_graph()](#metaknowledge.write_graph), if this dow not produce the desired results the networkx builtin [networkx.read_edgelist()](https://networkx.github.io/documentation/networkx-1.9.1/reference/generated/networkx.readwrite.edgelist.read_edgelist.html) could be tried.

    The read edge list format assumes the column named _eSource_ (From) is the source node, then the next column _eDest_ (To) givens the destination and all other columns are attributes of the edge, e.g. weight.

    The read nodeList format assumes the column called _idKey_ is the ID of the node as used by the edge list and the resulting network. All other columns are considered attributes of the node, e.g. count.

    If the names of the columns do not match those given to **read_graph()** a KeyError exception will be raised.

    **Note**: if nodes appear in the edgelist but not the nodeList they will be created with no attributes.

    # Parameters

    _edgeList_ : `str`

    > a string giving the path to the edge list file

    _nodeList_ : `optional [str]`

    > a string giving the path to the node list file

    _directed_ : `optional [bool]`

    > default `False`, if `True` the produced network is directed instead of undirected

    _idKey_ : `optional [str]`

    > default `"ID"`, the name of the ID column in the node list

    _eSource_ : `optional [str]`

    > default `"From"`, the name of the source column in the edge list

    _eDest_ : `optional [str]`

    > default `"To"`, the name of the destination column in the edge list

    # Returns

    `networkx Graph`

    > the Graph described by the files
    """
    if metaknowledge.VERBOSE_MODE:
        PBar = _ProgressBar(0, "Starting to reading graphs")
    else:
        PBar = None
    if directed:
        grph = nx.DiGraph()
    else:
        grph = nx.Graph()
    if nodeList:
        if PBar:
            PBar.updateVal(0, "Reading " + nodeList)
        f = open(nodeList)
        nFile = csv.DictReader(f)
        for line in nFile:
            vals = line
            ndID = vals[idKey]
            del vals[idKey]
            if len(vals) > 0:
                grph.add_node(ndID, attr_dict=vals)
            else:
                grph.add_node(ndID)
        f.close()
    if PBar:
        PBar.updateVal(.5, "Reading " + edgeList)
    f = open(edgeList)
    eFile = csv.DictReader(f)
    for line in eFile:
        vals = line
        eFrom = vals[eSource]
        eTo = vals[eDest]
        del vals[eSource]
        del vals[eDest]
        if len(vals) > 0:
            grph.add_edge(eFrom, eTo, attr_dict = vals)
        else:
            grph.add_edge(eFrom, eTo)
    if PBar:
        PBar.finish(str(len(grph.nodes())) + " nodes and " + str(len(grph.edges())) + " edges found")
    f.close()
    return grph

def write_graph(grph, name, edgeInfo = True, typing = False, suffix = 'csv', overwrite = True):
    """Writes both the edge list and the node attribute list of _grph_ to files starting with _name_.

    The output files start with _name_, the file type (edgeList, nodeAttributes) then if typing is True the type of graph (directed or undirected) then the suffix, the default is as follows:

    >> name_fileType.suffix

    Both files are csv's with comma delimiters and double quote quoting characters. The edge list has two columns for the source and destination of the edge, "From" and "To" respectively, then, if _edgeInfo_ is `True`, for each attribute of the node another column is created. The node list has one column call "ID" with the node ids used by networkx and all other columns are the node attributes.

    To read back these files use [read_graph()](#metaknowledge.read_graph) and to write only one type of lsit use [write_edgeList()](#metaknowledge.write_edgeList) or [write_nodeAttributeFile()](#metaknowledge.write_nodeAttributeFile).

    **Warning**: this function will overwrite files, if they are in the way of the output, to prevent this set _overwrite_ to `False`

    **Note**: If any nodes or edges are missing an attribute a `KeyError` will be raised.

    # Parameters

    _grph_ : `networkx Graph`

    > A networkx graph of the network to be written.

    _name_ : `str`

    > The start of the file name to be written, can include a path.

    _edgeInfo_ : `optional [bool]`

    > Default `True`, if `True` the the attributes of each edge are written to the edge list.

    _typing_ : `optional [bool]`

    > Default `False`, if `True` the directed ness of the graph will be added to the file names.

    _suffix_ : `optional [str]`

    > Default `"csv"`, the suffix of the file.

    _overwrite_ : `optional [bool]`

    > Default `True`, if `True` files will be overwritten silently, otherwise an `OSError` exception will be raised.
    """
    if metaknowledge.VERBOSE_MODE:
        PBar = _ProgressBar(0, "Writing the graph to two files starting with: " + name)
    else:
        PBar = None
    if typing:
        if isinstance(grph, nx.classes.digraph.DiGraph) or isinstance(grph, nx.classes.multidigraph.MultiDiGraph):
            grphType = "_directed"
        else:
            grphType = "_undirected"
    else:
        grphType = ''
    nameCompts = os.path.split(os.path.expanduser(os.path.normpath(name)))
    if nameCompts[0] == '' and nameCompts[1] == '':
        edgeListName = "edgeList"+ grphType + '.' + suffix
        nodesAtrName = "nodeAttributes"+ grphType + '.' + suffix
    elif nameCompts[0] == '':
        edgeListName = nameCompts[1] + "_edgeList"+ grphType + '.' + suffix
        nodesAtrName = nameCompts[1] + "_nodeAttributes"+ grphType + '.' + suffix
    elif nameCompts[1] == '':
        edgeListName = os.path.join(nameCompts[0], "edgeList"+ grphType + '.' + suffix)
        nodesAtrName = os.path.join(nameCompts[0], "nodeAttributes"+ grphType + '.' + suffix)
    else:
        edgeListName = os.path.join(nameCompts[0], nameCompts[1] + "_edgeList"+ grphType + '.' + suffix)
        nodesAtrName = os.path.join(nameCompts[0], nameCompts[1] + "_nodeAttributes"+ grphType + '.' + suffix)
    if not overwrite:
        if os.path.isfile(edgeListName):
            raise OSError(edgeListName+ " already exists")
        if os.path.isfile(nodesAtrName):
            raise OSError(nodesAtrName + " already exists")
    write_edgeList(grph, edgeListName, extraInfo = edgeInfo, _progBar = PBar)
    if PBar:
        PBar.jumpUp()
    write_nodeAttributeFile(grph, nodesAtrName, _progBar = PBar)
    if PBar:
        PBar.finish(str(len(grph.nodes())) + " nodes and " + str(len(grph.edges())) + " edges written to file")

def write_edgeList(grph, name, extraInfo = True, _progBar = None):
    """Writes an edge list of _grph_ at the destination _name_.

    The edge list has two columns for the source and destination of the edge, "From" and "To" respectively, then, if _edgeInfo_ is `True`, for each attribute of the node another column is created.

    **Note**: If any edges are missing an attribute `KeyError` will be raised.

    # Parameters

    _grph_ : `networkx Graph`

    > The graph to be written to _name_

    _name_ : `str`

    > The name of the file to be written

    _edgeInfo_ : `optional [bool]`

    > Default `True`, if `True` the attributes of each edge will be written
    """
    if _progBar:
        count = 0
        eMax = len(grph.edges(data = True))
        if isinstance(_progBar, _ProgressBar):
            _progBar.updateVal(0, "Writing edge list " + name)
        else:
            _progBar = _ProgressBar(0, "Writing edge list " + name)
    if len(grph.edges(data = True)) < 1:
        outFile = open(name, 'w')
        outFile.write('"From","To"\n')
        outFile.close()
        if _progBar:
            _progBar.updateVal(1, "Done edge list " + name + ", 0 edges written.")
    else:
        if extraInfo:
            csvDict = ['From'] +  ['To'] + list(grph.edges_iter(data = True).__next__()[2].keys())
        else:
            csvDict = ['From'] +  ['To']
        f = open(name, 'w')
        outFile = csv.DictWriter(f, csvDict, delimiter = ',', quotechar = '"', quoting=csv.QUOTE_ALL)
        outFile.writeheader()
        if extraInfo:
            for e in grph.edges_iter(data = True):
                if _progBar:
                    count += 1
                    _progBar.updateVal(count / eMax)
                eDict = e[2].copy()
                eDict['From'] = e[0]
                eDict['To'] = e[1]
                try:
                    outFile.writerow(eDict)
                except ValueError:
                    raise ValueError("Some edges in " + str(grph) + " do not have the same attributes")
        else:
            for e in grph.edges_iter():
                if _progBar:
                    count += 1
                    if count % 100 == 0:
                        _progBar.updateVal(count / eMax)
                eDict['From'] = e[0]
                eDict['To'] = e[1]
                outFile.writerow(eDict)
        if _progBar:
            _progBar.finish("Done edge list " + name + ", " + str(count) + " edges written.")
        f.close()

def write_nodeAttributeFile(grph, name, _progBar = None):
    """Writes a node attribute list of _grph_ with filename _name_

    The node list has one column call "ID" with the node ids used by networkx and all other columns are the node attributes.

    **Note**: If any edges are missing an attribute `KeyError` will be raised.

    # Parameters

    _grph_ : `networkx Graph`

    > The graph to be written to _name_

    _name_ : `str`

    > The name of the file to be written
    """
    if _progBar:
        count = 0
        nMax = len(grph.nodes())
        if isinstance(_progBar, _ProgressBar):
            _progBar.updateVal(0, "Writing edgelist " + name)
        else:
            _progBar = _ProgressBar(0, "Writing edgelist " + name)
    if len(grph.nodes(data = True)) < 1:
        outFile = open(name, 'w')
        outFile.write('ID\n')
        outFile.close()
        if _progBar:
            _progBar.updateVal(1, "Done node attribute list: " + name + ", 0 nodes written.")
    else:
        csvDict = ['ID'] + list(grph.nodes_iter(data = True).__next__()[1].keys())
        f = open(name, 'w')
        outFile = csv.DictWriter(f, csvDict, delimiter = ',', quotechar = '"', quoting=csv.QUOTE_ALL)
        outFile.writeheader()
        for n in grph.nodes_iter(data = True):
            if _progBar:
                count += 1
                _progBar.updateVal(count / nMax)
            nDict = n[1].copy()
            nDict['ID'] = n[0]
            try:
                outFile.writerow(nDict)
            except ValueError:
                raise ValueError("Some nodes in " + str(grph) + " do not have the same attributes")
        if _progBar:
            _progBar.updateVal(1, "Done node attribute list: " + name + ", " + str(count) + " nodes written.")
        f.close()

class _ProgressBar(object):
    difTermAndBar = 8 #the number of characters difference between the bar's length and the terminal's width
    timeLength = 6 # width of elapse time display
    percLength = 6 # width of percent display
    def __init__(self, initPer, initString = ' ', output = sys.stdout, secondRow = False):
        self.finished = False
        self.big = secondRow
        self.per = initPer
        self.out = output
        self.inputString = initString
        self.sTime = time.time()
        try:
            self.barMaxLength = os.get_terminal_size(self.out.fileno()).columns - self.difTermAndBar
            if self.barMaxLength < 0:
                self.barMaxLength = 0
        except OSError:
            self.barMaxLength = 80 - self.difTermAndBar
        self.ioThread = threading.Thread(target = self.threadedUpdate, kwargs = {"self" : self})
        self.ioThread.daemon = True
        self.ioThread.start()

    def __del__(self):
        if not self.finished:
            self.finished = True
            self.ioThread.join()
            self.out.write('\n\n')
            self.out.flush()

    def updateVal(self, inputPer, inputString = None):
        self.per = inputPer
        if inputString is not None:
            self.inputString = inputString

    def finish(self, inputString):
        self.finished = True
        self.inputString = str(inputString)
        self.ioThread.join()
        try:
            self.barMaxLength = os.get_terminal_size(self.out.fileno()).columns - self.difTermAndBar
            if self.barMaxLength < 0:
                self.barMaxLength = 0
        except OSError:
            self.barMaxLength = 80 - self.difTermAndBar
        if self.big:
            self.out.write('\n' + ' ' * (self.barMaxLength + self.difTermAndBar) + '\033[F')
        else:
            self.out.write('\r')
        if len(self.inputString) < self.barMaxLength + self.difTermAndBar - self.timeLength:
            tString = self.prepTime(time.time() - self.sTime, self.barMaxLength + self.difTermAndBar - len(self.inputString) - 1)
            self.out.write(self.inputString + ' ' + tString)
        else:
            self.out.write(self.prepString(self.inputString, self.barMaxLength + self.difTermAndBar - self.timeLength) + self.prepTime(time.time() - self.sTime, self.timeLength))
        self.out.write('\n')
        self.out.flush()

    def jumpUp(self):
        self.out.write('\033[F')
        self.out.flush()

    @staticmethod
    def prepString(s, maxLength):
        maxLength = maxLength - 1
        sString = str(s)
        if len(sString) <= maxLength:
            return sString.ljust(maxLength, ' ') + ' '
        else:
            if maxLength % 2 == 0:
                return sString[:int(maxLength/2 - 3)] + '...' + sString[int(-maxLength/2):] + ' '
            else:
                return sString[:int(maxLength/2 - 2)] + '...' + sString[int(-maxLength/2):] + ' '
    @staticmethod
    def prepTime(t, maxLength):
        try:
            if math.log10(t) + 3.01 > maxLength:
                return "{1:{0}.0E}s".format(maxLength - 1 ,t)
            else:
                return "{1:{0}.1f}s".format(maxLength - 1,t)
        except ValueError:
            return "{1:{0}.1f}s".format(maxLength - 1,t)

    @staticmethod
    def threadedUpdate(self = None):
        while not self.finished:
            try:
                self.barMaxLength = os.get_terminal_size(self.out.fileno()).columns - self.difTermAndBar
                if self.barMaxLength < 0:
                    self.barMaxLength = 0
            except OSError:
                self.barMaxLength = 80 - self.difTermAndBar
            self.out.write('\r')
            percentString = '{:.1%}'.format(self.per).rjust(self.percLength, ' ')
            barLength = int(self.per * self.barMaxLength)
            if self.big and self.inputString:
                    self.dString = self.prepString(self.inputString, self.barMaxLength + self.difTermAndBar - self.timeLength) + self.prepTime(time.time() - self.sTime, self.timeLength)
                    if barLength >= self.barMaxLength:
                        self.out.write('[' + '=' * barLength + ']' + percentString)
                        self.out.write('\n' + self.dString + '\033[F')
                    else:
                        self.out.write('[' + '=' * barLength + '>' + ' ' * (self.barMaxLength - barLength - 1) + ']' + percentString)
                        self.out.write('\n' + self.dString + '\033[')
            elif self.inputString:
                self.dString = self.prepString(self.inputString, self.barMaxLength + self.difTermAndBar - self.timeLength - self.percLength - 2) + '[' + self.prepTime(time.time() - self.sTime, self.timeLength) +  ']' + percentString
                self.out.write(self.dString)
            else:
                if barLength >= self.barMaxLength:
                    self.out.write('[' + '=' * barLength + ']' + percentString + '\r')
                else:
                    self.out.write('[' + '=' * barLength + '>' + ' ' * (self.barMaxLength - barLength - 1) + ']' + percentString + '\r')
            self.out.flush()
            time.sleep(.1)

def getWeight(grph, nd1, nd2, weightString = "weight", returnType = int):
    """
    A way of getting the weight of an edge with or without weight as a parameter
    returns a the value of the weight parameter converted to returnType if it is given or 1 (also converted) if not
    """
    if not weightString:
        return returnType(1)
    else:
        return returnType(grph.edge[nd1][nd2][weightString])

def getNodeDegrees(grph, weightString = "weight", strictMode = False,  returnType = int, edgeType = 'bi'):
    """
    Retunrs a dictionary of nodes to their degrees, the degree is determined by adding the weight of edge with the weight being the string weightString that gives the name of the attribute of each edge containng thier weight. The Weights are then converted to the type returnType. If weightString is give as False instead each edge is counted as 1.

    edgeType, takes in one of three strings: 'bi', 'in', 'out'. 'bi' means both nodes on the edge count it, 'out' mans only the one the edge comes form counts it and 'in' means only the node the edge goes to counts it. 'bi' is the default. Use only on directional graphs as otherwise the selected nodes is random.
    """
    ndsDict = {}
    for nd in grph.nodes_iter():
        ndsDict[nd] = returnType(0)
    for e in grph.edges_iter(data = True):
        if weightString:
            try:
                edgVal = returnType(e[2][weightString])
            except KeyError:
                if strictMode:
                    raise KeyError("The edge from " + str(e[0]) + " to " + str(e[1]) + " does not have the attribute: '" + str(weightString) + "'")
                else:
                    edgVal = returnType(1)
        else:
            edgVal = returnType(1)
        if edgeType == 'bi':
            ndsDict[e[0]] += edgVal
            ndsDict[e[1]] += edgVal
        elif edgeType == 'in':
            ndsDict[e[1]] += edgVal
        elif edgeType == 'out':
            ndsDict[e[0]] += edgVal
        else:
            raise ValueError("edgeType must be 'bi', 'in', or 'out'")
    return ndsDict

def getDegreeDistribution(grph, weightParameter = "weight", strictWeightNames = False,  weightType = int, directionalType = 'bi'):
    if weightType != int:
        raise ValueError("Unsupported type for weights, only ints are supported")
    ndsDict = getNodeDegrees(grph, weightString = weightParameter, strictMode = strictWeightNames, returnType = weightType, edgeType = directionalType)
    distVec = [0] * (max(ndsDict.values()) + 1)
    for v in ndsDict.values():
        distVec[v] += 1
    return distVec

def drop_edges(grph, minWeight = - float('inf'), maxWeight = float('inf'), parameterName = 'weight', ignoreUnweighted = False, dropSelfLoops = False):
    """
    Returns a graph with edges whose weight is within the inclusive bounds of minWeight and maxWeight, i.e minWeight <= edges weight <= maxWeight, will throw a Keyerror if the graph is unweighted

    minWeight and maxWeight default to negative and positive infinity respectively so without specifying either the output should be the input

    parameterName is key to weight field in the edge's dictionary, default is weight as that is almost always correct

    ignoreUnweighted can be set False to suppress the KeyError and make unweighted edges be ignored
    """
    if metaknowledge.VERBOSE_MODE:
        PBar = _ProgressBar(0, "Dropping edges")
        count = 0
        total = len(grph.edges())
    else:
        PBar = None
    tmpGrph = grph.copy()
    if minWeight != - float('inf') or maxWeight != float('inf'):
        for e in grph.edges_iter(data = True):
            try:
                val = e[2][parameterName]
            except KeyError:
                if not ignoreUnweighted:
                    raise KeyError("One or more Edges do not have weight or " + str(parameterName), " is not the name of the weight")
                else:
                    pass
            else:
                if PBar:
                    count += 1
                    if count % 100000 == 0:
                        PBar.updateVal(count/ total, str(count) + " edges analysed and " + str(total -len(tmpGrph.edges())) + " edges dropped")
                if val > maxWeight or  val < minWeight:
                    tmpGrph.remove_edge(e[0], e[1])
    if dropSelfLoops:
        slps = tmpGrph.selfloop_edges()
        if PBar:
            PBar.updateVal(1, "Dropping self {} loops".format(len(slps)))
        for e in slps:
            tmpGrph.remove_edge(e[0], e[1])
    if PBar:
        PBar.finish(str(total - len(tmpGrph.edges())) + " edges out of " + str(total) + " dropped, " + str(len(tmpGrph.edges())) + " returned")
    return tmpGrph

def drop_nodesByDegree(grph, minDegree = -float('inf'), maxDegree = float('inf'), useWeight = True, parameterName = 'weight', ignoreUnweighted = True):
    """
    Returns a graph whose nodes have a degree that is within inclusive bounds of minDegree and maxDegree, i.e minDegree <= degree <= maxDegree. Degree can be determined in two ways by default it is the total number of edges touching a node, alternative if useWeight is True it is the sum of the weight of all the edges touching a node.

    minDegree and maxDegree default to negative and positive infinity respectively so without specifying either the output should be the input

    useWeight can be set True to use an alternative method for calculating degree, the total weight of all edges

    parameterName is key to weight field in the edge's dictionary, default is weight as that is almost always correct, only used if useWeight is True

    ignoreUnweighted can be set False to suppress the KeyError and make unweighted edges be not counted, only used if useWeight is True
    """
    if metaknowledge.VERBOSE_MODE:
        PBar = _ProgressBar(0, "Dropping nodes by degree")
        count = 0
        total = len(grph.nodes())
    else:
        PBar = None
    goodNodes = []
    for n in grph.nodes_iter():
        if PBar:
            count += 1
            if count % 10000 == 0:
                PBar.updateVal(count/ total, str(count) + " nodes analysed and " + str(total - len(goodNodes)) + " nodes dropped")
        val = 0
        if useWeight:
            for e in grph.edges(n, data = True):
                try:
                    val += e[2][parameterName]
                except KeyError:
                    if not ignoreUnweighted:
                        raise KeyError("One or more Edges do not have weight or " + str(parameterName), " is not the name of the weight")
                    else:
                        val += 1
        else:
            val = len(grph.edges(n))
        if val <= maxDegree and val >= minDegree:
            goodNodes.append(n)
    if PBar:
        PBar.finish(str(total - len(goodNodes)) + " nodes out of " + str(total) + " dropped, " + str(len(goodNodes)) + " returned")
    return grph.subgraph(goodNodes)

def drop_nodesByCount(grph, minCount = -float('inf'), maxCount = float('inf'), parameterName = 'count', ignoreMissing = False):
    """
    Returns a graph whose nodes have a occurrence count that is within inclusive bounds of minCount and maxCount, i.e minCount <= count <= maxCount. Occurrence count is determined by reading the variable associated with the node named parameterName.

    minCount and maxCount default to negative and positive infinity respectively so without specifying either the output should be the input


    parameterName is key to count field in the node's dictionary, default is count as that is often correct

    ignoreMissing can be set False to suppress the KeyError and make nodes missing counts be dropped instead of throwing errors
    """
    if metaknowledge.VERBOSE_MODE:
        PBar = _ProgressBar(0, "Dropping nodes by count")
        count = 0
        total = len(grph.nodes())
    else:
        PBar = None
    goodNodes = []
    for n in grph.nodes_iter(data = True):
        if PBar:
            count += 1
            if count % 10000 == 0:
                PBar.updateVal(count/ total, str(count) + "nodes analysed and " + str(total - len(goodNodes)) + " nodes dropped")

        try:
            val = n[1][parameterName]
        except KeyError:
            if not ignoreMissing:
                raise KeyError("One or more nodes do not have counts or " + str(parameterName), " is not the name of the count parameter")
            else:
                pass
        else:
            if val <= maxCount and val >= minCount:
                goodNodes.append(n[0])
    if PBar:
        PBar.finish(str(total - len(goodNodes)) + " nodes out of " + str(total) + " dropped, " + str(len(goodNodes)) + " returned")
    return grph.subgraph(goodNodes)

def graphStats(G, stats = ('nodes', 'edges', 'isolates', 'loops', 'density', 'transitivity'), makeString = True):
    """Returns a string or list containing statistics about the graph _G_.

    **graphStats()** gives 6 different statistics: number of nodes, number of edges, number of isolates, number of loops, density and transitivity. The ones wanted can be given to _stats_. By default a string giving a sentence containing all the requested statistics is returned but the raw values can be accessed instead by setting _makeString_ to `False`.

    # Parameters

    _G_ : `networkx Graph`

    > The graph for the statistics to be determined of

    _stats_ : `optional [list or tuple [str]]`

    > Default 'nodes', 'edges', 'isolates', 'loops', 'density', 'transitivity'), a list or tuple containing any number or combination of the strings:

    > `"nodes"`, `"edges"`, `"isolates"`, `"loops"`, `"density"` and `"transitivity"``

    > At least one occurrence of the corresponding string causes the statistics to be provided in the string output. For the non-string (tuple) output the returned tuple has the same length as the input and each output is at the same index as the string that requested it, e.g.

    > `_stats_ = ("edges", "loops", "edges")`

    > The return is a tuple with 2 elements the first and last of which are the number of edges and the second is the number of loops

    _makeString_ : `optional [bool]`

    > Default `True`, if `True` a string is returned if `False` a tuple

    # Returns

    `str or tuple [float and int]`

    > The type is determined by _makeString_ and the layout by _stats_
     """

    for sts in stats:
        if sts not in ['nodes', 'edges', 'isolates', 'loops', 'density', 'transitivity']:
            raise RuntimeError('"{}" is not a valid stat.'.format(sts))
    if makeString:
        stsData = []
    else:
        stsData = {}
    if 'nodes' in stats:
        if makeString:
            stsData.append("{:G} nodes".format(len(G.nodes())))
        else:
            stsData['nodes'] = len(G.nodes())
    if 'edges' in stats:
        if makeString:
            stsData.append("{:G} edges".format(len(G.edges())))
        else:
            stsData['edges'] = len(G.edges())
    if 'isolates' in stats:
        if makeString:
            stsData.append("{:G} isolates".format(len(nx.isolates(G))))
        else:
            stsData['isolates'] = len(nx.isolates(G))
    if 'loops' in stats:
        if makeString:
            stsData.append("{:G} self loops".format(len(G.selfloop_edges())))
        else:
            stsData['loops'] = len(G.selfloop_edges())
    if 'density' in stats:
        if makeString:
            stsData.append("a density of {:G}".format(nx.density(G)))
        else:
            stsData['density'] = nx.density(G)
    if 'transitivity' in stats:
        if makeString:
            stsData.append("a transitivity of {:G}".format(nx.transitivity(G)))
        else:
            stsData['transitivity'] = nx.transitivity(G)
    if makeString:
        retString = "The graph has "
        if len(stsData) < 1:
            return retString
        elif len(stsData) == 1:
            return retString + stsData[0]
        else:
            return retString + ', '.join(stsData[:-1]) + ' and ' + stsData[-1]
    else:
        retLst = []
        for sts in stats:
            retLst.append(stsData[sts])
        return tuple(retLst)



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
import metaknowledge
from .record import Record, BadISIFile
from .graphHelpers import _ProgressBar
from .constants import tagsAndNames, tagToFull, fullToTag
from .citation import Citation

import itertools
import os.path
import csv

import networkx as nx

class RecordCollection(object):
    """
    A way of containing a large number of Record objects, it provides ways of creating them from an isi file, string, list of records or directory containing isi files. The Records are containing within a set and as such many of the set operations are defined, pop, union, in ... also records are hashed with their WOS string so no duplication can occur.
    The comparison operators <, <=, >, >= are based strictly on the number of Records within the collection, while equality looks for an exact match on the Records

    When being created if there are issues the Record collection will be declared bad (self.bad = True) it will then mostly return nothing or False. The error attribute contains the exception that occurred.

    They also possess a name accessed with repr(), this is used to auto generate the names of files and can be set at creation, note though that any operations that modify the RecordCollection's contents will update the name to include what occurred, read __repr__'s doc string for more information

    inCollection is the object containing the information about the Records to be constructed it can be an isi file, string, list of records or directory containing isi files

    name sets the name of the of the record if left blank name will be generated based on the object that created the Recordcollection

    extension controls the extension that __init__ looks for when reading a directory, set it to the extension on the isi files you wish to load, if left blank all files will be tried and any that are not isi files will be silently skipped

    # \_\_Init\_\_

    RecordCollections are made from either a single file or directory supplied as _inCollection_.

    # Parameters

    _inCollection_ : `optional [str] or None`

    > the name of the source of WOS records. It can be skipped to produce an empty collection.

    > If a file is provided. First it is checked to see if it is a WOS file (the header is checked). Then records are read from it one by one until the 'EF' string is found indicating the end of the file.

    > If a directory is provided. First each file in the directory is checked for the correct header and all those that do are then read like indivual files. The records are then collected into a single set in the RecordCollection.

    _name_ : `optional [str]`

    > The name of the RecordCollection, defaults to empty string. If left empty the name of the Record collection is set to the name of the file or directory used to create the collection. If provided the name id set to _name_

    _extension_ : `optional [str]`

    > The extension to search for when reading a directoy for files. _extension_ is the suffix searched for when a direcorty is read for files, by default it is empty so all files are read.
    """

    def __init__(self, inCollection = None, name = '', extension = ''):
        self.bad = False
        self._repr = name
        if not inCollection:
            if not name:
                self._repr = "empty"
            self._Records = set()
        elif isinstance(inCollection, str):
            if os.path.isfile(inCollection):
                try:
                    if not inCollection.endswith(extension):
                        raise TypeError("extension of input file does not match requested extension")
                    self._repr = os.path.splitext(os.path.split(inCollection)[1])[0]
                    self._Records = set(isiParser(inCollection))
                except BadISIFile as w:
                    self.bad = True
                    self.error = w
            elif os.path.isdir(inCollection):
                if metaknowledge.VERBOSE_MODE:
                    PBar = _ProgressBar(0, "Reading files from " + str(inCollection))
                    count = 0
                else:
                    PBar = None
                if extension and not name:
                    if extension[0] == '.':
                        self._repr = extension[1:] + "-files-from-" + os.path.dirname(inCollection)
                    else:
                        self._repr = extension + "-files-from-" + inCollection
                elif not name:
                    self._repr = "files-from-" + inCollection
                self._Records = set()
                flist = []
                for f in os.listdir(inCollection):
                    fullF = os.path.join(os.path.abspath(inCollection), f)
                    if fullF.endswith(extension) and os.path.isfile(fullF):
                        flist.append(fullF)
                for file in flist:
                    if PBar:
                        count += 1
                        PBar.updateVal(count / len(flist), "Reading records from: " + file)
                    else:
                        PBar = None
                    try:
                        self._Records |= set(isiParser(file))
                    except BadISIFile:
                        if extension != '':
                            raise
                        else:
                            pass
                    except UnicodeDecodeError:
                        pass
                if PBar:
                    PBar.finish("Done reading records from: " + str(inCollection))
            else:
                raise TypeError("inCollection is not a directory or a file")
        elif isinstance(inCollection, list):
            self._Records = set(inCollection)
        elif isinstance(inCollection, set):
            self._Records = inCollection
        else:
            raise TypeError

    def __add__(self, other):
        """
        returns the union of the two RecordCollections
        """
        if self.bad and other.bad:
            return RecordCollection(set(), '[BAD ' + repr(self) + ']'+ '_plus_' + '[BAD ' +  repr(other) + ']')
        if self.bad:
            return RecordCollection(other._Records, '[BAD ' + repr(self) + ']' + '_plus_' + repr(other))
        elif other.bad:
            return RecordCollection(self._Records,  repr(self) + '[BAD ' + repr(other) + ']')
        else:
            return RecordCollection(self._Records | other._Records, repr(self) + '_plus_' + repr(other))

    def __and__(self, other):
        """
        returns the intersection of the two RecordCollections
        """
        if self.bad or other.bad:
            raise Exception
        else:
            return RecordCollection(self._Records & other._Records, repr(self) + '_and_' + repr(other))

    def __sub__(self, other):
        """
        returns the difference of the two RecordCollections
        """
        if self.bad or other.bad:
            raise Exception
        else:
            return RecordCollection(self._Records - other._Records, repr(self) + '_diff_' + repr(other))

    def __xor__(self, other):
        """
        returns the symmetric difference of the two RecordCollections
        """
        if self.bad or other.bad:
            raise Exception
        else:
            return RecordCollection(self._Records ^ other._Records, repr(self) + '_symdiff_' + repr(other))

    def __str__(self):
        """
        Returns a string giving the the number Records in a sentence:
        "Collection of # records"
        """
        return "Collection of " + str(len(self._Records)) + " records"

    def __repr__(self):
        """
        The name of the RecordCollection, this used to identify the Collection when it is written as a file by writeFile() or in some of CL scripts
        It is updated when some modification of the RecordCollection occurs i.e.
        >>> RC = metaknowledge.RecordCollection('.', extension = 'isi')
        >>> repr(RC)
        isi-files-from-.
        >>> R = RC.pop()
        >>> repr(RC)
        pop-isi-files-from-.
        """
        return self._repr

    def __lt__(self, other):
        if self.bad or other.bad:
            return False
        else:
            return len(self) < len(other)
    def __le__(self, other):
        if self.bad or other.bad:
            return False
        else:
            return len(self) <= len(other)
    def __gt__(self, other):
        if self.bad or other.bad:
            return False
        else:
            return len(self) > len(other)
    def __ge__(self, other):
        if self.bad or other.bad:
            return False
        else:
            return len(self) >= len(other)

    def __eq__(self, other):
        if self.bad or other.bad:
            return False
        else:
            return self._Records == other._Records

    def __ne__(self, other):
        return not self == other

    def __len__(self):
        """
        returns the number or Records
        """
        return len(self._Records)

    def __iter__(self):
        """
        iterates over the Records
        """
        for R in self._Records:
            yield R

    def __contains__(self, item):
        """
        Returns True if the item is a WOS string of one of the Records containing or if item is a Record checks if the record is in the RecordCollection, otherwise returns False
        """
        if isinstance(item, str):
            for R in self:
                if R.UT == item:
                    return True
            return False
        elif isinstance(item, Record):
            return item in self._Records
        else:
            return False


    def pop(self):
        """
        Returns a random Record from the recordCollection, the Record is deleted from the collection, use peak for nondestructive access
        """
        if len(self._Records) > 0:
            self._repr = "Pop-" + self._repr
            return self._Records.pop()
        else:
            return None

    def peak(self):
        """
        Returns a random Record from the recordCollection, the Record is kept in the collection, use pop for faster destructive access
        """
        if len(self._Records) > 0:
            return self._Records.__iter__().__next__()
        else:
            return None

    def dropWOS(self, wosNum):
        """Removes the Record with WOS number (ID number) _wosNum_

        # Parameters

        _wosNum_ : `str`

        > _wosNum_ is the WOS number of the Record to be dropped. _wosNum_ must begin with 'WOS:' or a valueError is raise.
        """
        if wosNum[:4] != 'WOS:':
            raise ValueError("{} is not a valid WOS number string, it does not start with 'WOS:'.".format(wosNum))
        for R in self._Records:
            if R.wosString == wosNum:
                self._Records.remove(R)
                break

    def addRec(self, Rec):
        """Adds a Record or Records to the RecordCollection.

        # Parameters

        _Rec_ : `Record or iterable[Record]`

        > A Record or some iterable containg records to add
        """
        if hasattr(Rec, '__iter__'):
            self._Records |= set(Rec)
        elif Rec not in self:
            self._Records.add(Rec)

    def getWOS(self, wosNum, drop = False):
        """Gets the Record from the collection by its WOS number.

        # Parameters

        _wosNum_ : `str`

        > _wosNum_ is the WOS number of the Record to be extracted. _wosNum_ must begin with 'WOS:' or a valueError is raise.

        _drop_ : `optional [bool]`

        > Default `False`. If `True` the Record is dropped from the collection after being extract, i.e. if `False` [getWOS()](#RecordCollection.getWOS) acts like [peak()](#RecordCollection.peak), if `True` it acts like [pop()](#RecordCollection.pop)

        # Returns

        `metaknowledge.Record`

        > The Record whose WOS number is _wosNum_
        """
        try:
            if wosNum[:4] != 'WOS:':
                raise ValueError("{} is not a valid WOS number string, it does not start with 'WOS:'.".format(wosNum))
        except (TypeError, IndexError):
            raise ValueError("{} is not a valid WOS number string, it does not start with 'WOS:'.".format(wosNum))
        for R in self:
            if R.wosString == wosNum:
                if drop:
                    self._Records.remove(R)
                return R
        return None

    def getBadRecords(self):
        """
        returns RecordCollection containing all the Record which have their bad flag set to True, i.e. all those removed by dropBadRecords()
        """
        badRecords = set()
        for R in self._Records:
            if R.bad:
                badRecords.add(R)
        return RecordCollection(badRecords, repr(self) + '_badRecords')

    def dropBadRecords(self):
        """
        Removes all Records with bad attributes == True from the collection
        """
        self._Records = {r for r in self._Records if not r.bad}
        self._repr = repr(self) + '_badRecordsDropped'

    def dropNonJournals(self, ptVal = 'J', dropBad = True, invert = False):
        """Drops the non journal type Records from the collection

        # Parameters

        _ptVal_ : `optional [str]`

        > The value of the PT tag to be kept, default is 'J' the journal tag

        _dropBad_ : `optional [bool]`

        > Determines if bad Records will be dropped as well, default `True`

        _invert_ : `optional [bool]`

        > Set `True` to drop journals (or the PT tag given by _ptVal) instead of keeping them. Note, it still drops bad Records if _dropBad_ is `True`, default `False`
        """
        if dropBad:
            self.dropBadRecords()
        if invert:
            self._Records = {r for r in self._Records if r.PT != ptVal.upper()}
            self._repr = repr(self) + '_PT-{}-Dropped'.format(ptVal)
        else:
            self._Records = {r for r in self._Records if r.PT == ptVal.upper()}
            self._repr = repr(self) + '_PT-{}-Only'.format(ptVal)

    def writeFile(self, fname = None):
        """
        Writes the RecordCollection to a file, the written file is identical to those download from WOS. The order of Records written is random.

        fname set the name of the file, if blank the RecordCollection's name's first 200 characters are use with the suffix .isi
        """
        if fname:
            f = open(fname, mode = 'w', encoding = 'utf-8')
        else:
            f = open(repr(self)[:200] + '.isi', mode = 'w', encoding = 'utf-8')
        f.write("\ufeffFN Thomson Reuters Web of Science\u2122\n")
        f.write("VR 1.0\n")
        for R in self._Records:
            R.writeRecord(f)
            f.write('\n')
        f.write('EF')
        f.close()

    def writeCSV(self, fname = None, onlyTheseTags = None, longNames = False, firstTags = None, csvDelimiter = ',', csvQuote = '"', listDelimiter = '|'):
        """Writes all the Records from the collection into a csv file with each row a record and each column a tag

        fname is the name of the file to write to, if none is given it uses the Collections name suffixed by .csv

        onlyTheseTags lets you specify which tags to use, if not given then all tags in the records are given.
        If you want to use all known tags the use onlyTheseTags = metaknowledge.knownTagsList

        longNames if set to True will convert the tags to their longer names, otherwise the short 2 character ones will be used

        firstTags is the column's tags, it is set by default to ['UT', 'PT', 'TI', 'AF', 'CR'] so those will be written first if given by onlyTheseTags.
        Note if tags are in firstTags but not in onlyTheseTags, onlyTheseTags will override onlyTheseTags

        csvDelimiter is the delimiter used for the cells of the csv file, default is the comma (,)

        csvQuote is  the quote character used for the csv, default is the double quote (")

        listDelimiter is the delimiter used between values of the same cell if the tag for that record has multiple outputs, default is the pipe (|)
        """
        if firstTags is None:
            firstTags = ['UT', 'PT', 'TI', 'AF', 'CR']
        for i in range(len(firstTags)):
            if firstTags[i] in fullToTag:
                firstTags[i] = fullToTag[firstTags[i]]
        if onlyTheseTags:
            for i in range(len(onlyTheseTags)):
                if onlyTheseTags[i] in fullToTag:
                    onlyTheseTags[i] = fullToTag[onlyTheseTags[i]]
            retrievedFields = [t for t in firstTags if t in onlyTheseTags] + [t for t in onlyTheseTags if t not in firstTags]
        else:
            retrievedFields = firstTags
            for R in self:
                tagsLst = [t for t in R.activeTags() if t not in retrievedFields]
                retrievedFields += tagsLst
        if longNames:
            try:
                retrievedFields = [tagToFull[t] for t in retrievedFields]
            except KeyError:
                raise KeyError("One of the tags could not be converted to a long name.")
        if fname:
            f = open(fname, mode = 'w', encoding = 'utf-8')
        else:
            f = open(repr(self)[:200] + '.csv', mode = 'w', encoding = 'utf-8')
        csvWriter = csv.DictWriter(f, retrievedFields, delimiter = csvDelimiter, quotechar = csvQuote, quoting=csv.QUOTE_ALL)
        csvWriter.writeheader()
        for R in self:
            recDict = R.getTagsDict(retrievedFields)
            for k in recDict.keys():
                value = recDict[k]
                if hasattr(value, '__iter__'):
                    recDict[k] = listDelimiter.join([str(v) for v in value])
                elif recDict[k] == None:
                    recDict[k] = ''
                else:
                    recDict[k] = str(value)
            csvWriter.writerow(recDict)
        f.close()


    def makeDict(self, onlyTheseTags = None, longNames = False, cleanedVal = True):
        """Returns a dict with each key a tag and the values being lists of the values for each of the Records in the collection, `None` is given when there is no value and they are in the same order across each tag.

        When used in pandas: `pandas.DataFrame(RC.makeDict())` returns a data frame with each column a tag and each row a Record.

        # Parameters

        See writeCSV()
        """
        if onlyTheseTags:
            for i in range(len(onlyTheseTags)):
                if onlyTheseTags[i] in fullToTag:
                    onlyTheseTags[i] = fullToTag[onlyTheseTags[i]]
            retrievedFields = onlyTheseTags
        else:
            retrievedFields = []
            for R in self:
                tagsLst = [t for t in R.activeTags() if t not in retrievedFields]
                retrievedFields += tagsLst
        if longNames:
            try:
                retrievedFields = [tagToFull[t] for t in retrievedFields]
            except KeyError:
                raise KeyError("One of the tags could not be converted to a long name.")
        retDict = {k : [] for k in retrievedFields}
        for R in self:
            for k, v in R.getTagsDict(retrievedFields, cleaned = cleanedVal).items():
                retDict[k].append(v)
        return retDict

    def coAuthNetwork(self):
        """Creates a coauthorship network for the RecordCollection.

        # Returns

        `Networkx Graph`

        > A networkx graph with author names as nodes and collaborations as edges.
        """
        grph = nx.Graph()
        if metaknowledge.VERBOSE_MODE:
            PBar = _ProgressBar(0, "Starting to make a co-authorship network")
            pcount = 0
        else:
            PBar = None
        for R in self:
            if PBar:
                pcount += 1
                PBar.updateVal(pcount/ len(self), "Analyzing: " + str(R))
            authsList = R.authorsFull
            if authsList:
                authsList = list(authsList)
                if len(authsList) > 1:
                    for i, auth1 in enumerate(authsList):
                        if auth1 not in grph:
                            grph.add_node(auth1, count = 1)
                        else:
                            grph.node[auth1]['count'] += 1
                        for auth2 in authsList[i + 1:]:
                            if auth2 not in grph:
                                grph.add_node(auth2, count = 1)
                            else:
                                grph.node[auth2]['count'] += 1
                            if grph.has_edge(auth1, auth2):
                                grph.edge[auth1][auth2]['weight'] += 1
                            else:
                                grph.add_edge(auth1, auth2, weight = 1)
                else:
                    auth1 = authsList[0]
                    if auth1 not in grph:
                        grph.add_node(auth1, count = 1)
                    else:
                        grph.node[auth1]['count'] += 1
        if PBar:
            PBar.finish("Done making a co-authorship network")
        return grph

    def coCiteNetwork(self, dropAnon = True, nodeType = "full", nodeInfo = True, fullInfo = False, weighted = True, dropNonJournals = False, count = True, keyWords = None):
        """Creates a co-citation network for the RecordCollection.

        # Parameters

        _nodeType_ : `optional [str]`

        > One of `"full"`, `"original"`, `"author"`, `"journal"` or `"year"`. Specifies the value of the nodes in the graph. The default `"full"` causes the citations to be compared holistically using the [`metaknowledge.Citation`](#Citation.Citation) builtin comparison operators. `"original"` uses the raw original strings of the citations. While `"author"`, `"journal"` and `"year"` each use the author, journal and year respectively.

        _dropAnon_ : `optional [bool]`

        > default `True`, if `True` citations labeled anonymous are removed from the network

        _nodeInfo_ : `optional [bool]`

        > default `True`, wether an extra piece of information is stored with each node.

        _fullInfo_ : `optional [bool]`

        > default `False`, wether the original citation string is added to the node as an extra value, the attribute is labeled as fullCite

        _weighted_ : `optional [bool]`

        > default `True`, wether the edges are weighted. If `True` the edges are weighted by the number of citations.

        _dropNonJournals_ : `optional [bool]`

        > default `False`, wether to drop citations of non-journals

        _count_ : `optional [bool]`

        > default `True`, causes the number of occurrences of a node to be counted

        _keyWords_ : `optional [str] or [list[str]]`

        > A string or list of strings that the citations are checked against, if they contain any of the strings they are removed from the network

        # Returns

        `Networkx Graph`

        > A networkx graph with hashes as ID and co-citation as edges
        """
        allowedTypes = ["full", "original", "author", "journal", "year"]
        if nodeType not in allowedTypes:
            raise ValueError("{} is not an allowed nodeType.".format(nodeType))
        tmpgrph = nx.Graph()
        if metaknowledge.VERBOSE_MODE:
            PBar = _ProgressBar(0, "Starting to make a co-citation network")
            count = 0
        else:
            PBar = None
        for R in self:
            if PBar:
                count += 1
                PBar.updateVal(count / len(self), "Analyzing: " + str(R))
            Cites = R.citations
            if Cites:
                filteredCites = filterCites(Cites, nodeType, dropAnon, dropNonJournals, keyWords)
                addToNetwork(tmpgrph, filteredCites, count, weighted, nodeType, nodeInfo , fullInfo, headNd = None)
        if PBar:
            PBar.finish("Done making a co-citation network of " + repr(self))
        return tmpgrph


    def citationNetwork(self, dropAnon = True, nodeType = "full", nodeInfo = True, fullInfo = False, weighted = True, dropNonJournals = False, count = True, directed = True, keyWords = None):

        """Creates a citation network for the RecordCollection.

        # Parameters

        _nodeType_ : `optional [str]`

        > One of `"full"`, `"original"`, `"author"`, `"journal"` or `"year"`. Specifies the value of the nodes in the graph. The default `"full"` causes the citations to be compared holistically using the [`metaknowledge.Citation`](#Citation.Citation) builtin comparison operators. `"original"` uses the raw original strings of the citations. While `"author"`, `"journal"` and `"year"` each use the author, journal and year respectively.

        _dropAnon_ : `optional [bool]`

        > default `True`, if `True` citations labeled anonymous are removed from the network

        _nodeInfo_ : `optional [bool]`

        > default `True`, wether an extra piece of information is stored with each node.

        _fullInfo_ : `optional [bool]`

        > default `False`, wether the original citation string is added to the node as an extra value, the attribute is labeled as fullCite

        _weighted_ : `optional [bool]`

        > default `True`, wether the edges are weighted. If `True` the edges are weighted by the number of citations.

        _dropNonJournals_ : `optional [bool]`

        > default `False`, wether to drop citations of non-journals

        _count_ : `optional [bool]`

        > default `True`, causes the number of occurrences of a node to be counted

        _keyWords_ : `optional [str] or [list[str]]`

        > A string or list of strings that the citations are checked against, if they contain any of the strings they are removed from the network

        _directed_ : `optional [bool]`

        > Determines if the output graph is directed, default `True`

        # Returns

        `Networkx DiGraph or Networkx Graph`

        > See _directed_ for explanation of returned type

        > A networkx digraph with hashes as ID and citations as edges
        """
        allowedTypes = ["full", "original", "author", "journal", "year"]
        if nodeType not in allowedTypes:
            raise ValueError("{} is not an allowed nodeType.".format(nodeType))
        if directed:
            tmpgrph = nx.DiGraph()
        else:
            tmpgrph = nx.Graph()
        if metaknowledge.VERBOSE_MODE:
            PBar = _ProgressBar(0, "Starting to make a citation network")
            count = 0
        else:
            PBar = None
        for R in self:
            if PBar:
                count += 1
                PBar.updateVal(count/ len(self), "Analyzing: " + str(R))
            reRef = R.createCitation()
            if len(filterCites([reRef], nodeType, dropAnon, dropNonJournals, keyWords)) == 0:
                continue
            rCites = R.citations
            if rCites:
                filteredCites = filterCites(rCites, nodeType, dropAnon, dropNonJournals, keyWords)
                addToNetwork(tmpgrph, filteredCites, count, weighted, nodeType, nodeInfo , fullInfo, headNd = reRef)
        if PBar:
            PBar.finish("Done making a citation network of " + repr(self))
        return tmpgrph

    def _extractTagged(self, taglist):
        recordsWithTags = set()
        for R in self:
            for t in taglist:
                hasTags = True
                if t not in R.tags:
                    hasTags = False
                    break
            if hasTags:
                recordsWithTags.add(R)
        return RecordCollection(recordsWithTags, repr(self) + "_tags(" + ','.join(taglist) + ')')

    def yearSplit(self, startYear, endYear):
        """Creates a RecordCollection of Records from the years between _startYear_ and _endYear_ inclusive.

        # Parameters

        _startYear_ : `int`

        > The smallest year to be included in the retuned RecordCollection

        _endYear_ : `int`

        > The largest year to be included in the retuned RecordCollection

        # Returns

        `RecordCollection`

        > A RecordCollection of Records from _startYear_ to _endYear_
        """

        recordsInRange = set()
        for R in self._Records:
            if R.year >= startYear and R.year <= endYear:
                recordsInRange.add(R)
        return RecordCollection(recordsInRange, repr(self) + "_(" + str(startYear) + " ," + str(endYear) + ")")

    def oneModeNetwork(self, mode, nodeCount = True, edgeWeight = True):
        """Creates a network of the objects found by one WOS tag _mode_.

        A **oneModeNetwork()** looks are each Record in the RecordCollection and extracts its values for the tag given by _mode_, e.g. the `"AF"` tag. Then if multiple are returned an edge is created between them. So in the case of the author tag `"AF"` a co-authorship network is created.

        The number of times each object occurs is count if _nodeCount_ is `True` and the edges count the number of co-occurrences if _edgeWeight_ is `True`. Both are`True` by default.

        **Note** Do not use this for the construction of co-citation networks use [Recordcollection.coCiteNetwork()](#RecordCollection.coCiteNetwork) it is more accurate and has more options.

        # Parameters

        _mode_ : `str`

        > A two character WOS tag or one of the full names for a tag

        _nodeCount_ : `optional [bool]`

        > Default `True`, if `True` each node will have an attribute called "count" that contains an int giving the number of time the object occurred.

        _edgeWeight_ : `optional [bool]`

        > Default `True`, if `True` each edge will have an attribute called "weight" that contains an int giving the number of time the two objects co-occurrenced.

        # Returns

        `networkx Graph`

        > A networkx Graph with the objects of the tag _mode_ as nodes and their co-occurrences as edges
        """
        if mode not in tagsAndNames:
            raise TypeError(str(mode) + " is not a known tag, or the name of a known tag.")
        if metaknowledge.VERBOSE_MODE:
            PBar = _ProgressBar(0, "Starting to make a one mode network with " + mode)
            count = 0
        else:
            PBar = None
        grph = nx.Graph()
        for R in self:
            if PBar:
                count += 1
                PBar.updateVal(count / len(self), "Analyzing: " + str(R))
            contents = getattr(R, mode, None)
            if contents:
                if isinstance(contents, list):
                    tmplst = [str(n) for n in contents]
                    if len(tmplst) > 1:
                        for i, node1 in enumerate(tmplst):
                            for node2 in tmplst[i + 1:]:
                                if edgeWeight:
                                    try:
                                        grph.edge[node1][node2]['weight'] += 1
                                    except KeyError:
                                        grph.add_edge(node1, node2, weight = 1)
                                else:
                                    if not grph.has_edge(node1, node2):
                                        grph.add_edge(node1, node2)
                            if nodeCount:
                                try:
                                    grph.node[node1]['count'] += 1
                                except KeyError:
                                    grph.node[node1]['count'] = 1
                            else:
                                if not grph.has_node(node1):
                                    grph.add_node(node1)
                    elif len(tmplst) == 1:
                        if nodeCount:
                            try:
                                grph.node[tmplst[0]]['count'] += 1
                            except KeyError:
                                grph.add_node(tmplst[0], count = 1)
                        else:
                            if not grph.has_node(tmplst[0]):
                                grph.add_node(tmplst[0])
                    else:
                        pass
                else:
                    nodeVal = str(contents)
                    if nodeCount:
                        try:
                            grph.node[nodeVal]['count'] += 1
                        except KeyError:
                            grph.add_node(nodeVal, count = 1)
                    else:
                        if not grph.has_node(nodeVal):
                            grph.add_node(nodeVal)
        if PBar:
            PBar.finish("Done making a one mode network with " + mode)
        return grph

    def twoModeNetwork(self, tag1, tag2, directed = False, recordType = True, nodeCount = True, edgeWeight = True):
        """Creates a network of the objects found by two WOS tags _tag1_ and _tag2_.

        A **twoModeNetwork()** looks are each Record in the RecordCollection and extracts its values for the tags given by _tag1_ and _tag2_, e.g. the `"WC"` and `"LA"` tags. Then for each object returned by each tag and edge is created between it and every other object of the other tag. So the WOS defined subject tag `"WC"` and language tag `"LA"`, will give a two-mode network showing the connections between subjects and languages. Each node will have an attribute call `"type"` that gives the tag that created it or both if both created it, e.g. the node `"English"` would have the type attribute be `"LA"`.

        The number of times each object occurs is count if _nodeCount_ is `True` and the edges count the number of co-occurrences if _edgeWeight_ is `True`. Both are`True` by default.

        The _directed_ parameter if `True` will cause the network to be directed with the first tag as the source and the second as the destination.

        # Parameters

        _mode_ : `str`

        > A two character WOS tag or one of the full names for a tag

        _directed_ : `optional [bool]`

        > Default `False`, if `True` the returned network is directed

        _nodeCount_ : `optional [bool]`

        > Default `True`, if `True` each node will have an attribute called "count" that contains an int giving the number of time the object occurred.

        _edgeWeight_ : `optional [bool]`

        > Default `True`, if `True` each edge will have an attribute called "weight" that contains an int giving the number of time the two objects co-occurrenced.

        # Returns

        `networkx Graph or networkx DiGraph`

        > A networkx Graph with the objects of the tags _tag1_ and _tag2_ as nodes and their co-occurrences as edges.
        """
        if (not tag1 in tagsAndNames) or (not tag2 in tagsAndNames):
            raise TypeError(str(tag1) + " or " + str(tag2) + " is not a known tag, or the name of a known tag.")
        if metaknowledge.VERBOSE_MODE:
            PBar = _ProgressBar(0, "Starting to make a two mode network of " + tag1 + " and " + tag2)
            count = 0
        else:
            PBar = None
        if directed:
            grph = nx.DiGraph()
        else:
            grph = nx.Graph()
        for R in self:
            if PBar:
                count += 1
                PBar.updateVal(count / len(self), "Analyzing: " + str(R))
            contents1 = getattr(R, tag1, None)
            contents2 = getattr(R, tag2, None)
            if isinstance(contents1, list):
                contents1 = [str(v) for v in contents1]
            elif contents1 == None:
                contents1 = []
            else:
                contents1 = [str(contents1)]
            if isinstance(contents2, list):
                contents2 = [str(v) for v in contents2]
            elif contents2 == None:
                contents2 = []
            else:
                contents2 = [str(contents2)]
            for node1 in contents1:
                for node2 in contents2:
                    if edgeWeight:
                        try:
                            grph.edge[node1][node2]['weight'] += 1
                        except KeyError:
                            grph.add_edge(node1, node2, weight = 1)
                    else:
                        if not grph.has_edge(node1, node2):
                            grph.add_edge(node1, node2)
                if nodeCount:
                    try:
                        grph.node[node1]['count'] += 1
                    except KeyError:
                        try:
                            grph.node[node1]['count'] = 1
                            if recordType:
                                grph.node[node1]['type'] = tag1
                        except KeyError:
                            if recordType:
                                grph.add_node(node1, type = tag1)
                            else:
                                grph.add_node(node1)
                else:
                    if not grph.has_node(node1):
                        if recordType:
                            grph.add_node(node1, type = tag1)
                        else:
                            grph.add_node(node1)
                    elif recordType:
                        if 'type' not in grph.node[node1]:
                            grph.node[node1]['type'] = tag1

            for node2 in contents2:
                if nodeCount:
                    try:
                        grph.node[node2]['count'] += 1
                    except KeyError:
                        try:
                            grph.node[node2]['count'] = 1
                            if recordType:
                                grph.node[node2]['type'] = tag2
                        except KeyError:
                            grph.add_node(node2, count = 1)
                            if recordType:
                                grph.node[node2]['type'] = tag2
                else:
                    if not grph.has_node(node2):
                        if recordType:
                            grph.add_node(node2, type = tag2)
                        else:
                            grph.add_node(node2)
                    elif recordType:
                        if 'type' not in grph.node[node2]:
                            grph.node[node2]['type'] = tag2
        if PBar:
            PBar.finish("Done making a two mode network of " + tag1 + " and " + tag2)
        return grph

    def nModeNetwork(self, tags, recordType = True, nodeCount = True, edgeWeight = True):
        """Creates a network of the objects found by all WOS tags in _tags_.

        A **nModeNetwork()** looks are each Record in the RecordCollection and extracts its values for the tags given by _tags_. Then for all objects returned an edge is created between them, regardless of their type. Each node will have an attribute call `"type"` that gives the tag that created it or both if both created it, e.g. if `"LA"` were in _tags_ node `"English"` would have the type attribute be `"LA"`.

        For example if _tags_ was set to `['CR', 'UT', 'LA']`, a three mode network would be created, composed of a co-citation network from the `"CR"` tag. Then each citation would also have edges to all the languages of Records that cited it and to the WOS number of the those Records.

        The number of times each object occurs is count if _nodeCount_ is `True` and the edges count the number of co-occurrences if _edgeWeight_ is `True`. Both are`True` by default.

        # Parameters

        _mode_ : `str`

        > A two character WOS tag or one of the full names for a tag

        _nodeCount_ : `optional [bool]`

        > Default `True`, if `True` each node will have an attribute called "count" that contains an int giving the number of time the object occurred.

        _edgeWeight_ : `optional [bool]`

        > Default `True`, if `True` each edge will have an attribute called "weight" that contains an int giving the number of time the two objects co-occurrenced.

        # Returns

        `networkx Graph`

        > A networkx Graph with the objects of the tags _tags_ as nodes and their co-occurrences as edges
        """
        for t in tags:
            if t not in tagsAndNames:
                raise TypeError(str(t) + " is not a known tag, or the name of a known tag.")
        if metaknowledge.VERBOSE_MODE:
            PBar = _ProgressBar(0, "Starting to make a " + str(len(tags)) + "-mode network of: " + ', '.join(tags))
            count = 0
        else:
            PBar = None
        grph = nx.Graph()
        for R in self:
            if PBar:
                count += 1
                PBar.updateVal(count / len(self), "Analyzing: " + str(R))
            contents = []
            for t in tags:
                tmpVal = getattr(R, t, None)
                if tmpVal:
                    if isinstance(tmpVal, list):
                        contents.append((t, [str(v) for v in tmpVal]))
                    else:
                        contents.append((t, [str(tmpVal)]))
            for i, vlst1 in enumerate(contents):
                for node1 in vlst1[1]:
                    for vlst2 in contents[i + 1:]:
                        for node2 in vlst2[1]:
                            if edgeWeight:
                                try:
                                    grph.edge[node1][node2]['weight'] += 1
                                except KeyError:
                                    grph.add_edge(node1, node2, weight = 1)
                            else:
                                if not grph.has_edge(node1, node2):
                                    grph.add_edge(node1, node2)
                    if nodeCount:
                        try:
                            grph.node[node1]['count'] += 1
                        except KeyError:
                            try:
                                grph.node[node1]['count'] = 1
                                if recordType:
                                    grph.node[node1]['type'] = vlst1[0]
                            except KeyError:
                                if recordType:
                                    grph.add_node(node1, type = vlst1[0])
                                else:
                                    grph.add_node(node1)
                    else:
                        if not grph.has_node(node1):
                            if recordType:
                                grph.add_node(node1, type = vlst1[0])
                            else:
                                grph.add_node(node1)
                        elif recordType:
                            try:
                                grph.node[node1]['type'] += vlst1[0]
                            except KeyError:
                                grph.node[node1]['type'] = vlst1[0]
        if PBar:
            PBar.finish("Done making a " + str(len(tags)) + "-mode network of: " +  ', '.join(tags))
        return grph

    def localCiteStats(self, pandasFriendly = False, keyType = "citation"):
        """Returns a dict with all the citations in the CR field as keys and the number of times they occur as the values

        # Parameters

        _pandasFriendly_ : `optional [bool]`

        > default `False`, makes the output be a dict with two keys one "Citations" is the citations the other is their occurence counts as "Counts".

        _keyType_ : `optional [str]`

        > default `'citation'`, the type of key to use for the dictionary, the valid strings are `"citation"`, `"journal"`, `"year"` or `"author"`

        # Returns

        `dict[str, int or Citataion : int]`

        > A dictioanry with keys as given by _keyType_ and integers giving their rates of occurnce in the collection
        """
        if metaknowledge.VERBOSE_MODE:
            PBar = _ProgressBar(0, "Starting to get the local stats on {}s.".format(keyType))
            count = 0
            recCount = len(self)
        else:
            PBar = None
        keyTypesLst = ["citation", "journal", "year", "author"]
        citesDict = {}
        if keyType not in keyTypesLst:
            raise TypeError("{} is not a valid key type, only '{}' or '{}' are.".format(keyType, "', '".join(keyTypesLst[:-1], keyTypesLst[-1]) ))
        for R in self:
            rCites = R.CR
            if PBar:
                count += 1
                PBar.updateVal(count / recCount, "Analysing: {}".format(R.UT))
            if rCites:
                for c in rCites:
                    if keyType == keyTypesLst[0]:
                        cVal = c
                    else:
                        cVal = getattr(c, keyType)
                        if cVal is None:
                            continue
                    if cVal in citesDict:
                        citesDict[cVal] += 1
                    else:
                        citesDict[cVal] = 1
        if PBar:
            PBar.finish("Done, {} {} fields analysed".format(len(citesDict), keyType))
        if pandasFriendly:
            citeLst = []
            countLst = []
            for cite, occ in citesDict.items():
                citeLst.append(cite)
                countLst.append(occ)
            return {"Citations" : citeLst, "Counts" : countLst}
        else:
            return citesDict

    def localCitesOf(self, rec):
        """Takes in a Record, WOS string, citation string or Citation and returns a RecordCollection of all records that cite it.
        """
        localCites = []
        if isinstance(rec, Record):
            recCite = rec.createCitation()
        if isinstance(rec, str):
            try:
                recCite = self.getWOS(rec)
            except ValueError:
                try:
                    recCite = Citation(rec)
                except AttributeError:
                    raise ValueError("{} is not a valid WOS string or a valid citation string".format(recCite))
            else:
                if recCite is None:
                    return RecordCollection(inCollection = localCites, name = "Records_citing_{}".format(rec))
                else:
                    recCite = recCite.createCitation()
        elif isinstance(rec, Citation):
            recCite = rec
        else:
            raise ValueError("{} is not a vaild input, rec must be a Record, string or Citation object.".format(rec))
        for R in self:
            rCites = R.CR
            if rCites:
                for cite in rCites:
                    if recCite == cite:
                        localCites.append(R)
                        break
        return RecordCollection(inCollection = localCites, name = "Records_citing_'{}'".format(rec))

    def citeFilter(self, keyString = '', field = 'all', reverse = False, caseSensitive = False):
        """
        Filters Records by some string, keyString, in all of their citations.
        Returns all Records with at least one citation possessing keyString in the field given by field.

        keyString give the string to be searched for if it is is blank then all citations with the specified field will be matched

        field give the component of the citation to be looked at, it is one of a few strings. The default is 'all' which will cause the entire original citation to be searched. It can be used to search across fields, e.g. '1970, V2' is a valid keystring
        The other options are:
        author, searches the author field
        year, searches the year field
        journal, searches the journal field
        V, searches the volume field
        P, searches the page field
        misc, searches all the remaining uncategorized information
        anonymous, searches for anonymous citations, keyString is not used
        bad, searches for bad citations, keyString is not used

        reverse being True causes all Records not matching the query to be returned, default is False

        caseSensitive if True causes the search across the original to be case sensitive, only the 'all' option can be case sensitive

        """
        retRecs = []
        keyString = str(keyString)
        for R in self:
            try:
                if field == 'all':
                    for cite in R.citations:
                        if caseSensitive:
                            if keyString in cite.original:
                                retRecs.append(R)
                                break
                        else:
                            if keyString.upper() in cite.original.upper():
                                retRecs.append(R)
                                break
                elif field == 'author':
                    for cite in R.citations:
                        try:
                            if keyString.upper() in cite.author.upper():
                                retRecs.append(R)
                                break
                        except AttributeError:
                            pass
                elif field == 'journal':
                    for cite in R.citations:
                        try:
                            if keyString.upper() in cite.journal:
                                retRecs.append(R)
                                break
                        except AttributeError:
                            pass
                elif field == 'year':
                    for cite in R.citations:
                        try:
                            if int(keyString) == cite.year:
                                retRecs.append(R)
                                break
                        except AttributeError:
                            pass
                elif field == 'V':
                    for cite in R.citations:
                        try:
                            if keyString.upper() in cite.V:
                                retRecs.append(R)
                                break
                        except AttributeError:
                            pass
                elif field == 'P':
                    for cite in R.citations:
                        try:
                            if keyString.upper() in cite.P:
                                retRecs.append(R)
                                break
                        except AttributeError:
                            pass
                elif field == 'misc':
                    for cite in R.citations:
                        try:
                            if keyString.upper() in cite.misc:
                                retRecs.append(R)
                                break
                        except AttributeError:
                            pass
                elif field == 'anonymous':
                    for cite in R.citations:
                        if cite.isAnonymous():
                            retRecs.append(R)
                            break
                elif field == 'bad':
                    for cite in R.citations:
                        if cite.bad:
                            retRecs.append(R)
                            break
            except TypeError:
                pass
        if reverse:
            excluded = []
            for R in self:
                if R not in retRecs:
                    excluded.append(R)
            return RecordCollection(inCollection = excluded, name = self._repr + '_subsetByNotCite')
        else:
            return RecordCollection(inCollection = retRecs, name = self._repr + '_subsetByCite')


def isiParser(isifile):
    """
    isiParser() reads the file given by the path isifile, checks that the header is correct then reads until it reaches EF.
    Each it finds is used to initialize a Record then all Record are returned as a list.
    """
    try:
        openfile = open(isifile, 'r', encoding='utf-8-sig')
    except UnicodeDecodeError as e:
        openfile.close()
        raise e
    f = enumerate(openfile, start = 0)
    try:
        linesChecked = 3
        for i in range(linesChecked):
            if "VR 1.0" in f.__next__()[1]:
                break
            if i == linesChecked - 1:
                openfile.close()
                raise BadISIFile(isifile + " Does not have a valid header, 'VR 1.0' not in first two lines")
    except StopIteration as e:
        openfile.close()
        raise BadISIFile("File ends before EF found")
    except UnicodeDecodeError as e:
        openfile.close()
        raise e
    notEnd = True
    plst = []
    while notEnd:
        try:
            line = f.__next__()
        except StopIteration as e:
            raise BadISIFile("File ends before EF found")
        if not line[1]:
            raise BadISIFile("No ER found in " + isifile)
        elif line[1].isspace():
            continue
        elif 'EF' in line[1][:2]:
            notEnd = False
            continue
        else:
            try:
                plst.append(Record(itertools.chain([line], f), isifile, line[0]))
            except BadISIFile as w:
                try:
                    s = f.__next__()[1]
                    while s[:2] != 'ER':
                        s = f.__next__()[1]
                except:
                    raise BadISIFile(str(w) + " could not be resolved")
            except Exception as e:
                openfile.close()
                raise e
    try:
        f.__next__()
        raise BadISIFile("EF not at end of " + isifile)
    except StopIteration as e:
        pass
    finally:
        openfile.close()
    return plst

def getCoCiteIDs(clst):
    """
    Creates a dict of the ID-extra information pairs for a CR tag.
    """
    idDict = {}
    for c in clst:
        cId = c.getID()
        if cId not in idDict:
            idDict[cId] = c.getExtra()
    return idDict

def updateWeightedEdges(grph, ebunch):
    for e in ebunch:
        try:
            grph.edge[e[0]][e[1]]['weight'] += e[2]
        except KeyError:
            grph.add_edge(e[0], e[1], weight = e[2])

def edgeBunchGenerator(base, nodes, weighted = False, reverse = False):
    """
    A helper function for generating a bunch of edges from 1 node base to a list of nodes nodes.
    """
    if weighted and reverse:
        for n in nodes:
            yield (n, base, 1)
    elif weighted:
        for n in nodes:
            yield (base, n, 1)
    elif reverse:
        for n in nodes:
            yield (n, base)
    else:
        for n in nodes:
            yield (base, n)

def edgeNodeReplacerGenerator(base, nodes, loc):
    """
    A helper function for replacing an element of nodes at loc with base
    """
    for n in nodes:
        tmpN = list(n)
        tmpN[loc] = base
        yield tmpN


def addToNetwork(grph, nds, count, weighted, nodeType, nodeInfo , fullInfo, headNd = None):
    """Addeds the citations _nds_ to _grph_, arroring to the rules give by _nodeType_, _fullInfo_, etc.

    _headNd_ is the citation of the Record
    """
    if headNd is not None:
        hID = makeID(headNd, nodeType)
        if hID not in grph:
            grph.add_node(*makeNodeTuple(headNd, hID, nodeInfo, fullInfo, nodeType, count))
    else:
        hID = None
    idList = []
    for n in nds:
        nID = makeID(n, nodeType)
        if nID not in grph:
            grph.add_node(*makeNodeTuple(n, nID, nodeInfo, fullInfo, nodeType, count))
        elif count:
            grph.node[nID]['count'] += 1
        idList.append(nID)
    addedEdges = []
    if hID:
        for nID in idList:
            if weighted:
                try:
                    grph[hID][nID]['weight'] += 1
                except KeyError:
                    grph.add_edge(hID, nID, weight = 1)
            elif nID not in grph[hID]:
                addedEdges.append((hID, nID))
    elif len(idList) > 1:
        for i, outerID in enumerate(idList):
            for innerID in idList[i + 1:]:
                if weighted:
                    try:
                        grph[outerID][innerID]['weight'] += 1
                    except KeyError:
                        grph.add_edge(outerID, innerID, weight = 1)
                elif innerID not in grph[outerID]:
                    addedEdges.append((outerID, innerID))
    grph.add_edges_from(addedEdges)

def makeID(citation, nodeType):
    """Makes the id, of the correct type for the network"""
    if nodeType != "full":
        return getattr(citation, nodeType)
    else:
        return citation.getID()

def makeNodeTuple(citation, idVal, nodeInfo, fullInfo, nodeType, count):
    """Makes a tuple of idVal and a dict of the selected attributes"""
    d = {}
    if nodeInfo:
        if nodeType == 'full':
            d['info'] = str(citation)
        elif nodeType == 'journal':
            if citation.isJournal():
                d['info'] = str(citation.getFullJournalName())
            else:
                d['info'] = "None"
        elif nodeType == 'original':
            d['info'] = str(citation)
        else:
            d['info'] = idVal
    if fullInfo:
        d['fullCite'] = str(citation)
    if count:
        d['count'] = 1
    return (idVal, d)

def filterCites(cites, nodeType, dropAnon, dropNonJournals, keyWords):
    filteredCites = []
    for c in cites:
        if nodeType != "full" and not getattr(c, nodeType):
            pass
        elif dropNonJournals and not c.isJournal():
            pass
        elif dropAnon and c.isAnonymous():
            pass
        elif keyWords:
            found = False
            citeString = str(c).upper()
            if isinstance(keyWords, str):
                if keyWords.upper() in citeString:
                    found = True
            else:
                for k in keyWords:
                    if k.upper() in citeString:
                        found = True
                        break
            if not found:
                filteredCites.append(c)
        else:
            filteredCites.append(c)
    return filteredCites



"""The functions used by metaknowledge to handle WOS tags

Each of the functions in tagFuncs is named after the long name of its tag and is responsible for taking the raw data from a WOS file and returning usable information. The raw data is a list containing each line associated with the tag as a string. So the section of a record:

    TI The Motion Behind the Symbols: A Vital Role for Dynamism in the
      Conceptualization of Limits and Continuity in Expert Mathematics

would be the list:

    ["The Motion Behind the Symbols: A Vital Role for Dynamism in the",
    "Conceptualization of Limits and Continuity in Expert Mathematics"
    ]

The function to process it is called [`title()`](#tagFuncs.title) which is determined by looking up the tag in the `tagFuncs.tagToFunc` dictionary. Which is a dictionary mapping WOS tag strings to their functions. For a simple mapping of tags to their long strings use [`metaknowledge.tagToFull`](#metaknowledge.metaknowledge).

The full list of tags and their long names is provided below followed by the descriptions of the functions:

| tag | Name |
|:---|:---|
| `'EM'` |[email]({{ site.baseurl }}{{ page.url }}#tagFuncs.email)
| `'BS'` |[seriesSubtitle]({{ site.baseurl }}{{ page.url }}#tagFuncs.seriesSubtitle)
| `'VL'` |[volume]({{ site.baseurl }}{{ page.url }}#tagFuncs.volume)
| `'SO'` |[journal]({{ site.baseurl }}{{ page.url }}#tagFuncs.journal)
| `'PM'` |[pubMedID]({{ site.baseurl }}{{ page.url }}#tagFuncs.pubMedID)
| `'RP'` |[reprintAddress]({{ site.baseurl }}{{ page.url }}#tagFuncs.reprintAddress)
| `'PU'` |[publisher]({{ site.baseurl }}{{ page.url }}#tagFuncs.publisher)
| `'GP'` |[group]({{ site.baseurl }}{{ page.url }}#tagFuncs.group)
| `'J9'` |[j9]({{ site.baseurl }}{{ page.url }}#tagFuncs.j9)
| `'CR'` |[citations]({{ site.baseurl }}{{ page.url }}#tagFuncs.citations)
| `'EP'` |[endingPage]({{ site.baseurl }}{{ page.url }}#tagFuncs.endingPage)
| `'C1'` |[authAddress]({{ site.baseurl }}{{ page.url }}#tagFuncs.authAddress)
| `'UT'` |[wosString]({{ site.baseurl }}{{ page.url }}#tagFuncs.wosString)
| `'Z9'` |[totalTimesCited]({{ site.baseurl }}{{ page.url }}#tagFuncs.totalTimesCited)
| `'ID'` |[keyWords]({{ site.baseurl }}{{ page.url }}#tagFuncs.keyWords)
| `'PY'` |[year]({{ site.baseurl }}{{ page.url }}#tagFuncs.year)
| `'RI'` |[ResearcherIDnumber]({{ site.baseurl }}{{ page.url }}#tagFuncs.ResearcherIDnumber)
| `'EI'` |[eISSN]({{ site.baseurl }}{{ page.url }}#tagFuncs.eISSN)
| `'BF'` |[bookAuthorFull]({{ site.baseurl }}{{ page.url }}#tagFuncs.bookAuthorFull)
| `'BN'` |[ISBN]({{ site.baseurl }}{{ page.url }}#tagFuncs.ISBN)
| `'BP'` |[beginningPage]({{ site.baseurl }}{{ page.url }}#tagFuncs.beginningPage)
| `'AR'` |[articleNumber]({{ site.baseurl }}{{ page.url }}#tagFuncs.articleNumber)
| `'SU'` |[supplement]({{ site.baseurl }}{{ page.url }}#tagFuncs.supplement)
| `'NR'` |[citedRefsCount]({{ site.baseurl }}{{ page.url }}#tagFuncs.citedRefsCount)
| `'LA'` |[language]({{ site.baseurl }}{{ page.url }}#tagFuncs.language)
| `'BA'` |[bookAuthor]({{ site.baseurl }}{{ page.url }}#tagFuncs.bookAuthor)
| `'SN'` |[ISSN]({{ site.baseurl }}{{ page.url }}#tagFuncs.ISSN)
| `'CA'` |[groupName]({{ site.baseurl }}{{ page.url }}#tagFuncs.groupName)
| `'D2'` |[bookDOI]({{ site.baseurl }}{{ page.url }}#tagFuncs.bookDOI)
| `'PA'` |[publisherAddress]({{ site.baseurl }}{{ page.url }}#tagFuncs.publisherAddress)
| `'TC'` |[wosTimesCited]({{ site.baseurl }}{{ page.url }}#tagFuncs.wosTimesCited)
| `'OI'` |[orcID]({{ site.baseurl }}{{ page.url }}#tagFuncs.orcID)
| `'GA'` |[documentDeliveryNumber]({{ site.baseurl }}{{ page.url }}#tagFuncs.documentDeliveryNumber)
| `'AU'` |[authorsShort]({{ site.baseurl }}{{ page.url }}#tagFuncs.authorsShort)
| `'PG'` |[pageCount]({{ site.baseurl }}{{ page.url }}#tagFuncs.pageCount)
| `'SI'` |[specialIssue]({{ site.baseurl }}{{ page.url }}#tagFuncs.specialIssue)
| `'WC'` |[subjects]({{ site.baseurl }}{{ page.url }}#tagFuncs.subjects)
| `'CL'` |[confLocation]({{ site.baseurl }}{{ page.url }}#tagFuncs.confLocation)
| `'SP'` |[confSponsors]({{ site.baseurl }}{{ page.url }}#tagFuncs.confSponsors)
| `'PT'` |[pubType]({{ site.baseurl }}{{ page.url }}#tagFuncs.pubType)
| `'AB'` |[abstract]({{ site.baseurl }}{{ page.url }}#tagFuncs.abstract)
| `'BE'` |[editedBy]({{ site.baseurl }}{{ page.url }}#tagFuncs.editedBy)
| `'PI'` |[publisherCity]({{ site.baseurl }}{{ page.url }}#tagFuncs.publisherCity)
| `'HO'` |[confHost]({{ site.baseurl }}{{ page.url }}#tagFuncs.confHost)
| `'DI'` |[DOI]({{ site.baseurl }}{{ page.url }}#tagFuncs.DOI)
| `'AF'` |[authorsFull]({{ site.baseurl }}{{ page.url }}#tagFuncs.authorsFull)
| `'TI'` |[title]({{ site.baseurl }}{{ page.url }}#tagFuncs.title)
| `'SC'` |[subjectCategory]({{ site.baseurl }}{{ page.url }}#tagFuncs.subjectCategory)
| `'FX'` |[fundingText]({{ site.baseurl }}{{ page.url }}#tagFuncs.fundingText)
| `'DT'` |[docType]({{ site.baseurl }}{{ page.url }}#tagFuncs.docType)
| `'CT'` |[confTitle]({{ site.baseurl }}{{ page.url }}#tagFuncs.confTitle)
| `'IS'` |[issue]({{ site.baseurl }}{{ page.url }}#tagFuncs.issue)
| `'PD'` |[month]({{ site.baseurl }}{{ page.url }}#tagFuncs.month)
| `'JI'` |[isoAbbreviation]({{ site.baseurl }}{{ page.url }}#tagFuncs.isoAbbreviation)
| `'CY'` |[confDate]({{ site.baseurl }}{{ page.url }}#tagFuncs.confDate)
| `'DE'` |[authKeyWords]({{ site.baseurl }}{{ page.url }}#tagFuncs.authKeyWords)
| `'FU'` |[funding]({{ site.baseurl }}{{ page.url }}#tagFuncs.funding)
| `'PN'` |[partNumber]({{ site.baseurl }}{{ page.url }}#tagFuncs.partNumber)
| `'ED'` |[editors]({{ site.baseurl }}{{ page.url }}#tagFuncs.editors)
| `'SE'` |[seriesTitle]({{ site.baseurl }}{{ page.url }}#tagFuncs.seriesTitle)
| `'MA'` |[meetingAbstract]({{ site.baseurl }}{{ page.url }}#tagFuncs.meetingAbstract)

"""

from .tagFuncs import *



import os
import sys

def isInteractive():
    """
    A basic check of if the program is running in interactive mode
    """
    if sys.stdout.isatty():
        if os.name != 'nt':
            return True
        else:
            return False
    else:
        return False

VERBOSE_MODE = isInteractive()

#maps three letter combinations to numbers for use in Record.month
monthDict = {'SPR': 3, 'SUM': 6, 'FAL': 9, 'WIN': 12, 'JAN' : 1, 'FEB' : 2, 'MAR' : 3, 'APR' : 4, 'MAY' : 5, 'JUN' : 6 , 'JUL' : 7, 'AUG' : 8, 'SEP' : 9, 'OCT' : 10, 'NOV' : 11, 'DEC' : 12}

def makeBiDirectional(d):
    """
    Helper for generating tagNameConverter
    Makes dict that maps from key to value and back
    """
    dTmp = d.copy()
    for k in d:
        dTmp[d[k]] = k
    return dTmp

def reverseDict(d):
    """
    Helper for generating fullToTag
    Makes dict of value to key
    """
    retD = {}
    for k in d:
        retD[d[k]] = k
    return retD

"""
Converts WOS tags to more descriptive names
This is all tags known about and is mostly likely not a complete list
The names are not official and are added by the author
"""
tagToFull = {

            'PT' : "pubType",
            'AF' : "authorsFull",
            #'GA' : "groupAuthors", Old usage
            'GP' : "group",
            'BE' : "editedBy",
            'AU' : "authorsShort",
            'BA' : "bookAuthor",
            'BF' : "bookAuthorFull",
            'CA' : "groupName",
            'TI' : "title",
            'ED' : "editors",
            'SO' : "journal",
            'SE' : "seriesTitle",
            'BS' : "seriesSubtitle",
            'LA' : "language",
            'DT' : "docType",
            'CT' : "confTitle",
            'CY' : "confDate",
            'HO' : "confHost",
            'CL' : "confLocation",
            'SP' : "confSponsors",
            'DE' : "authKeyWords",
            'ID' : "keyWords",
            'AB' : "abstract",
            'C1' : "authAddress",
            'RP' : "reprintAddress",
            'EM' : "email",
            'RI' : "ResearcherIDnumber",
            'OI' : "orcID",
            'FU' : "funding",
            'FX' : "fundingText",
            'CR' : "citations",
            'NR' : "citedRefsCount",
            'TC' : "wosTimesCited",
            'Z9' : "totalTimesCited",
            'PU' : "publisher",
            'PI' : "publisherCity",
            'PA' : "publisherAddress",
            'SC' : "subjectCategory",
            'SN' : "ISSN",
            'EI' : "eISSN",
            'BN' : "ISBN",
            'J9' : "j9",
            'JI' : "isoAbbreviation",
            'PD' : "month",
            'PY' : "year",
            'VL' : "volume",
            'IS' : "issue",
            'PN' : "partNumber",
            'SU' : "supplement",
            'SI' : "specialIssue",
            'MA' : "meetingAbstract",
            'BP' : "beginningPage",
            'EP' : "endingPage",
            'AR' : "articleNumber",
            'PG' : "pageCount",
            'WC' : "subjects",
            'DI' : "DOI",
            'D2' : "bookDOI",
            'GA' : "documentDeliveryNumber",
            'UT' : "wosString",
            'PM' : "pubMedID",
            }

fullToTag = reverseDict(tagToFull) #Reverses tagToFull

tagNameConverter = makeBiDirectional(tagToFull) #tagToFull made reversible

tagsAndNames = set(tagNameConverter.keys()) #set of WOS tags ad their names

knownTagsList = list(tagToFull.keys()) #list of all the known tags



import metaknowledge
from .graphHelpers import _ProgressBar

import copy

#Based on Fast unfolding of communities in large networks http://arxiv.org/abs/0803.0476v2

def blondel(G, weightParameter = None, communityParameter = 'community'):
    if metaknowledge.VERBOSE_MODE:
        PBar = _ProgressBar(0, "Starting community detection")
        iterations = 1
        count = 0
    else:
        PBar = None
    communityDict = {}
    workingGrph = G.copy()
    for i, n in enumerate(G.nodes_iter()):
        workingGrph.node[n][communityParameter] = i
        communityDict[i] = [n]
    Qprevious = -2
    Qcurrent = modularity(workingGrph, weightParameter, communityParameter)
    communityToNode = copy.deepcopy(communityDict)
    while Qcurrent > Qprevious:
        m = get_m(G, weightParameter)
        newGrph = workingGrph.copy()
        for n, nDict in workingGrph.nodes_iter(data = True):
            if PBar:
                PBar.updateVal(count/len(workingGrph), "Processing sweep {0}, node {1} modularity is {2:.2f}".format(iterations, str(n), Qcurrent))
                count +=1
            current_com = nDict[communityParameter]
            currentMax = (0, current_com)
            for new_com in communityDict.keys():
                if new_com != current_com:
                    swapVal = deltaQswap(newGrph, n, new_com, communityDict, weightParameter, communityParameter, mVal = m)
                    if swapVal > currentMax[0]:
                        currentMax = (swapVal, new_com)
            if currentMax[0] > 0:
                newGrph.node[n][communityParameter] = currentMax[1]
                communityDict[current_com].remove(n)
                communityDict[currentMax[1]].append(n)
                if len(communityDict[current_com]) < 1:
                    del communityDict[current_com]
                try:
                    communityToNode[currentMax[1]] += communityToNode[current_com]
                    del communityToNode[current_com]
                except KeyError:
                    pass
        if PBar:
            PBar.updateVal(1, "Trimming sweep {0}, Q = {1:.1%}".format(iterations, Qcurrent))
            count = 0
            iterations += 1
        for com, nds in communityDict.items():
            if len(nds) > 1:
                for nd in nds[1:]:
                    merge_nodes(newGrph, nds[0], nd)
                communityDict[com] = [nds[0]]
        Qprevious = Qcurrent
        Qcurrent = modularity(newGrph, weightParameter, communityParameter)
        workingGrph = newGrph
    workingGrph = G.copy()
    if PBar:
        PBar.updateVal(1, "Done, updating graph")
    comNumList = list(communityToNode.keys())
    for ndcom, ndlst in communityToNode.items():
        for n in ndlst:
            workingGrph.node[n][communityParameter] = str(comNumList.index(ndcom))
    if PBar:
        PBar.finish("Done detecting communities in " + str(iterations) + " sweeps")
    return workingGrph

def modularity(G, weightParameter = None, communityParameter = 'community'):
    """
    Gets modularity of network, currently not tuned
    """
    m = 0
    kktot = 0
    Atot = 0
    for n_i in G.nodes_iter():
        for n_j in G.nodes_iter():
            A_ij = get_edgeVal(G, n_i, n_j, weightParameter)
            m += .5 * A_ij
            if G.node[n_i][communityParameter] == G.node[n_j][communityParameter]:
                Atot += A_ij
                kktot += get_kVal(G, n_i, weightParameter) * get_kVal(G, n_j, weightParameter)
    return ((Atot - 0.5 * kktot / m) / ( 2 * m))


def deltaQ(G, n, communitylst, weightParameter = 'weight', mCurrent = None):
    """
    Gets the change in modularity of a network by adding a single node, currently not tuned
    """
    if mCurrent:
        inversm = .5 / mCurrent
    else:
        inversm = .5 / get_m(G, weightParameter)
    Sigma_in = get_CommunityWeight(G, communitylst, weightParameter, internal = True)
    Sigma_tot = get_CommunityWeight(G, communitylst, weightParameter, internal = False)
    k_i = get_kVal(G, n, weightParameter)
    k_i_in = 0
    for e in G.edges_iter(communitylst, data = True):
        if e[1] == n:
            if weightParameter:
                k_i_in += e[2][weightParameter]
            else:
                k_i_in += 1
    dQ = Sigma_in + k_i_in - (Sigma_tot + k_i) * (Sigma_tot + k_i) * inversm
    dQ -= Sigma_in - Sigma_tot * Sigma_tot * inversm - k_i * k_i * inversm
    return dQ * inversm

def deltaQswap(G, n, newCom, comDict, weightParameter, communityParameter, mVal = None):
    oldCom = G.node[n][communityParameter]
    oldComlst = comDict[oldCom].copy()
    oldComlst.remove(n)
    oldDQ = deltaQ(G, n, oldComlst, weightParameter, mCurrent = mVal)
    newDQ = deltaQ(G, n, comDict[newCom], weightParameter, mCurrent = mVal)
    return newDQ - oldDQ

def merge_nodes(G, n1, n2):
    G.add_edges_from((n1, e[1], e[2]) for e in G.edges_iter(n2, data = True))
    G.remove_node(n2)

def get_m(G, weightParameter):
    m = 0
    if weightParameter:
        for e in G.edges_iter(data = True):
            m += e[2][weightParameter]
    else:
        return len(G.edges())
    return m

def get_kVal(G, n, weightParameter):
    """
    Helper function to get the total weight of all edges to a node
    """
    if weightParameter:
        retk = 0
        for e in G.edges_iter(n, data = True):
            retk += e[2][weightParameter]
    else:
        retk = len(G.edges(n))
    return retk


def get_CommunityWeight(G, community, weightParameter, internal = False):
    retTot = 0
    for e in G.edges_iter(community, data = True):
        if internal:
            if e[0] in community and e[1] in community:
                if weightParameter:
                    retTot += e[2][weightParameter]
                else:
                    retTot += 1
        else:
            if weightParameter:
                retTot += e[2][weightParameter]
            else:
                retTot += 1
    return retTot

def get_edgeVal(G, n1, n2, weightParameter):
    """
    Helper function to obtain the weight of the edge between two nodes, with exception handling, returns 0 if no edge is found
    """
    if weightParameter:
        try:
            valDict = G.edge[n1][n2]
        except KeyError:
            return 0
        else:
            return valDict[weightParameter]
    else:
        return 1



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
"""This file contains the Record class for metaknowledge and one helper function for parsing WOS records, recordParser. The record class is used to represent a single records meta-data from WOS.
"""
import itertools
import io
import collections

from .citation import Citation
from .constants import tagNameConverter, tagsAndNames, fullToTag
from .tagFuncs import tagToFunc

class BadISIRecord(Warning):
    """Exception thrown by the [record parser](#metaknowledge.recordParser) to indicate a mis-formated record. This occurs when some component of the record does not parse. The messages will be any of:

        * _Missing field on line (line Number):(line)_, which indicates a line was to short, there should have been a tag followed by information

        * _End of file reached before ER_, which indicates the file ended before the 'ER' indicator appeared, 'ER' indicates the end of a record. This is often due to a copy and paste error.

        * _Duplicate tags in record_, which indicates the record had 2 or more lines with the same tag.

        * _Missing WOS number_, which indicates the record did not have a 'UT' tag.

    Records with a BadISIRecord error are likely incomplete or the combination of two or more single records.
    """
    pass

class BadISIFile(Warning):
    """
    Exception thrown by isiParser for mis-formated files
    """
    pass

class Record(object):
    """
    Class for full WOS records

    It is meant to be immutable; many of the methods and attributes are evaluated when first called, not when the object is created, and the results are stored in a private dictionary.

    The record's meta-data is stored in an ordered dictionary labeled by WOS tags. To access the raw data stored in the original record the [getTag()](#Record.getTag) method can be used. To access data that has been processed and cleaned the attributes named after the tags are used.

    # Customizations

    The Record's hashing and equality testing are based on the WOS number (the tag is 'UT', and also called the accession number). They are strings starting with "WOS:" and followed by 15 or so numbers and letters, although both the length and character set are known to vary. The numbers are unique to each record so are used for comparisons. If a record is `bad`  all equality checks return `False`.

    When converted to a string the records title is used so for a record `R`, R.TI == R.title == str(R).

    # Attributes

    When a record is created if the parsing of the WOS file failed it is marked as `bad`. The `bad` attribute is set to True and the `error` attribute is created to contain the exception object.

    Generally, to get the information from a Record its attributes should be used. For a Record `R`, calling `R.CR` causes [citations()](#tagFuncs.citations) from the the [tagFuncs](#tagFuncs.tagFuncs) module to be called on the contents of the raw 'CR' field. Then the result is saved and returned. In this case, a list of Citation objects is returned. You can also call `R.citations` to get the same effect, as each known field tag has a longer name (currently there are 61 field tags). These names are meant to make accessing tags more readable and mapping from tag to name can be found in the tagToFull dict. If a tag is known (in [tagToFull](#metaknowledge.metaknowledge)) but not in the raw data `None` is returned instead. Most tags when cleaned return a string or list of strings, the exact results can be found in the help for the particular function.

    The attribute `authors` is also defined as a convience and returns the same as 'AF' or if that is not found 'AU'.

    # \_\_Init\_\_

    Records are generally create as collections in  [Recordcollections](#RecordCollection.RecordCollection), and not as individual objects. If you wish to create one on its own it is possible, the arguments are as follows.

    # Parameters

    _inRecord_: `files stream, dict, str or itertools.chain`

    > If it is a file stream the file must be open at the location of the first tag in the record, usually 'PT', and the file will be read until 'ER' is found, which indicates the end of the record in the file.

    > If a dict is passed the dictionary is used as the database of fields and tags, so each key is considered a WOS tag and each value a list of the lines of the original associated with the tag. This is the same form of dict that [recordParser](#metaknowledge.recordParser) returns.

    > For a str the input is the raw textual data of a single record in the WOS style, like the file stream it must start at the first tag and end in 'ER'.

    > itertools.chain is treated identically to a file stream and is used by [RecordCollections](#RecordCollection.RecordCollection).

    _sFile_ : `optional [str]`

    > Is the name of the file the raw data was in, by default it is blank. It is mostly used to make error messages more informative.

    _sLine_ : `optional [int]`

    > Is the line the record starts on in the raw data file. It is mostly used to make error messages more informative.
    """

    def __init__(self, inRecord, taglist = (), sFile = "", sLine = 0):
        """See help on [Record](#Record.Record) for details"""
        self._unComputedTags = set()
        self.bad = False
        self.error = None
        self.tags = taglist
        self._sourceFile = sFile
        self._sourceLine = sLine
        if isinstance(inRecord, dict):
            self._fieldDict = inRecord
        elif isinstance(inRecord, itertools.chain):
            try:
                self._fieldDict = recordParser(inRecord)
            except BadISIRecord as b:
                self.bad = True
                self.error = b
                self._fieldDict = {}
            finally:
                if hasattr(self, '_fieldDict') and 'UT' in self._fieldDict:
                    self._wosNum = self._fieldDict['UT'][0]
                else:
                    self._wosNum = None
                    self.bad = True
                    self.error = BadISIRecord("Missing WOS number")
        elif isinstance(inRecord, io.IOBase):
            try:
                self._fieldDict = recordParser(enumerate(inRecord))
            except BadISIRecord as b:
                self.bad = True
                self.error = b
                self._fieldDict = {}
            finally:
                if 'UT' in self._fieldDict:
                    self._wosNum = self._fieldDict['UT'][0]
                else:
                    self._wosNum = None
                    self.bad = True
                    self.error = BadISIRecord("Missing WOS number")
        elif isinstance(inRecord, str):
            try:
                def addChartoEnd(lst):
                    for s in lst:
                        yield s + '\n'
                self._fieldDict = recordParser(enumerate(addChartoEnd(inRecord.split('\n')), start = 1))
                #string io
            except BadISIRecord as b:
                self.bad = True
                self.error = b
                self._fieldDict = {}
            finally:
                if 'UT' in self._fieldDict:
                    self._wosNum = self._fieldDict['UT'][0]
                else:
                    self._wosNum = "NO WOS NUMBER"
                    self.bad = True
                    self.error = BadISIRecord("Missing WOS number")
        if hasattr(self, "_fieldDict"):
            for tag in self._fieldDict:
                if tag != 'UT':
                    self.__dict__[tag] = None
                    self._unComputedTags.add(tag)
                    try:
                        fullName = tagNameConverter[tag]
                    except KeyError:
                        pass
                    else:
                        self.__dict__[fullName] = None
                        self._unComputedTags.add(fullName)

    def __getattribute__(self, name):
        """
        Hack to get the attributes correct, read the Record help for more information
        """
        try:
            val = object.__getattribute__(self, name)
        except AttributeError:
            if name in tagsAndNames:
                return None
            else:
                raise
        else:
            if val != None:
                return val
            else:
                if name in self._unComputedTags:
                    try:
                        otherName = tagNameConverter[name]
                    except KeyError:
                        try:
                            tagVal = tagToFunc[name](self._fieldDict[name])
                        except KeyError:
                            tagVal = self._fieldDict[name]
                        setattr(self, name, tagVal)
                        self._unComputedTags.remove(name)
                    else:
                        try:
                            prossFunc = tagToFunc[name]
                        except KeyError:
                            try:
                                prossFunc = tagToFunc[otherName]
                            except KeyError:
                                prossFunc = lambda x: x
                        try:
                            tagVal = prossFunc(self._fieldDict[name])
                        except KeyError:
                            tagVal = prossFunc(self._fieldDict[otherName])
                        object.__setattr__(self, name, tagVal)
                        object.__setattr__(self, otherName, tagVal)
                        self._unComputedTags.remove(name)
                        self._unComputedTags.remove(otherName)
                return object.__getattribute__(self, name)


    def __str__(self):
        """
        returns a string with the title of the file as given by self.title(), if there is not one it returns "Untitled record"
        """
        if self.title:
            return self.title
        else:
            return "Untitled record"

    def __eq__(self, other):
        """
        returns true if the WOS numbers of both Records are identical.

        if either is bad False is returned
        """
        if not isinstance(other, Record):
            raise RuntimeError("Equality checking between Records and non-Records is not implemented")
        if self.bad or other.bad:
            return False
        else:
            return self.wosString == other.wosString

    def __ne__(self, other):
        """returns the opposite of \_\_eq\_\_"""
        return not self == other

    def __hash__(self):
        """returns a hash of the WOS number or if `bad` returns a hash of the fields combined with the error messages, either of these could be blank

        `bad` Records are more likely to cause hash collisions due to their lack of entropy when created.
        """
        if self.bad:
            return hash(str(self._fieldDict.values()) + str(self.error))
        return hash(self._wosNum)

    def __getstate__(self):
        """gets the __dict__ of the Record"""
        return self.__dict__

    def __setstate__(self, state):
        """This is necessary because __getattribute__ is overwritten, but works like \_\_setstate\_\_ usually does"""
        for k in state:
            object.__setattr__(self, k, state[k])

    @property
    def wosString(self):
        """Returns the WOS number (UT tag) of the record"""
        return self._wosNum

    @property
    def UT(self):
        """Returns the UT tag (WOS number) of the record"""
        return self._wosNum

    @property
    def authors(self):
        """Returns the full names of the authors (AF tag) if available otherwise attempts to use their shorter names (AU tag) if that fails None is returned
        Usually another way of calling authorsFull or AF
        """
        auth = self.authorsFull
        if auth is None:
            return self.authorsShort
        else:
            return auth

    def getTag(self, tag, clean = False):
        """Returns a list containing the raw data of the record associated with _tag_. Each line of the record is one string in the list.

        # Parameters

        _tag_ : `str`

        > _tag_ can be a two character string corresponding to a WOS tag e.g. 'J9', the matching is case insensitive so 'j9' is the same as 'J9'. Or it can be one of the full names for a tag with the mappings in [fullToTag](#metaknowledge). If the string is not found in the original record or after being translated through [fullToTag](#metaknowledge), `None` is returned.

        # Returns

        `List [str]`

        > Each string in the list is a line from the record associated with _tag_ or None if not found.
        """

        #tag = tag.upper()
        #TODO Figure out why this causes issues
        if clean:
            return getattr(self, tag.upper(), None)
        if tag in self._fieldDict:
            return self._fieldDict[tag]
        elif tag in fullToTag and fullToTag[tag] in self._fieldDict:
            return self._fieldDict[fullToTag[tag]]
        else:
            return None

    def createCitation(self):
        """Creates a citation string, using the same format as other WOS citations, for the [Record](#Record.Record) by reading the relevant tags (year, J9, volume, beginningPage, DOI) and using it to start a [Citation](#Citation.Citation) object.

        # Returns

        `Citation`

        > A [Citation](#Citation.Citation) object containing a citation for the Record.
        """
        valsLst = []
        if self.authorsShort:
            valsLst.append(self.authorsShort[0].replace(',', ''))
        if getattr(self, "year", False):
            valsLst.append(str(self.year))
        if getattr(self, "j9", False):
            valsLst.append(self.j9)
        elif getattr(self, "TI", False):
            valsLst.append(self.TI)
        if getattr(self, "volume", False):
            valsLst.append('V' + str(self.volume))
        if getattr(self, "beginningPage", False):
            valsLst.append('P' + str(self.beginningPage))
        if getattr(self, "DOI", False):
            valsLst.append('DOI ' + self.DOI)
        return Citation(', '.join(valsLst))

    def getTagsList(self, taglst, cleaned = False):
        """Returns a list of the results of [`getTag()`](#Record.getTag) for each tag in _taglist_, the return has the same order as the original.

        # Parameters
        _taglst_ : `List[str]`

        > Each string in _taglst_ can be a two character string corresponding to a WOS tag e.g. 'J9', the matching is case insensitive so 'j9' is the same as 'J9'. Or it can be one of the full names for a tag with the mappings in [fullToTag](#metaknowledge.metaknowledge). If the string is not found in the original record before or after being translated through [fullToTag](#metaknowledge.metaknowledge), `None` is used instead. Same as in [`getTag()`](#Record.getTag)

        > Then they are compiled into a list in the same order as _taglst_

        # Returns

        `List[str]`

        > a list of the values for each tag in _taglst_, in the same order
        """
        retList = []
        for tag in taglst:
            retList.append(self.getTag(tag), clean = cleaned)
        return retList

    def getTagsDict(self, taglst, cleaned = False):
        """returns a dict of the results of getTag, with the elements of _taglst_ as the keys and the results as the values.

        # Parameters
        _taglst_ : `List[str]`

        > Each string in _taglst_ can be a two character string corresponding to a WOS tag e.g. 'J9', the matching is case insensitive so 'j9' is the same as 'J9'. Or it can be one of the full names for a tag with the mappings in [fullToTag](#metaknowledge.metaknowledge). If the string is not found in the oriagnal record before or after being translated through [fullToTag](#metaknowledge), `None` is used instead. Same as in [`getTag()`](#Record.getTag)

        # Returns

        `dict[str : List [str]]`

        > a dictionary with keys as the original tags in _taglst_ and the values as the results
        """
        retDict = {}
        for tag in taglst:
            retDict[tag] = self.getTag(tag, clean = cleaned)
        return retDict

    def activeTags(self):
        """Returns a list of all the tags the original WOS record had. These are all the tags that ['getTag()'](#Record.getTag) will not return `None` for.

        # Returns

        `List[str]`

        > a list of WOS tags in the Record
        """
        return list(self._fieldDict.keys())

    def writeRecord(self, infile):
        """Writes to _infile_ the original contents of the Record. This is intended for use by [RecordCollections](#RecordCollection.RecordCollection) to write to file. What is written to _infile_ is bit for bit identical to the original record file. No newline is inserted above the write but the last character is a newline.

        # Parameters

        _infile_ : `file stream`

        > An open utf-8 encoded file
        """
        if self.bad:
            raise Exception
        else:
            for tag in self._fieldDict.keys():
                for i, value in enumerate(self._fieldDict[tag]):
                    if i == 0:
                        infile.write(tag + ' ')
                    else:
                        infile.write('   ')
                    infile.write(value + '\n')
            infile.write("ER\n")

def recordParser(paper):
    """Reads the file _paper_ until it reaches 'ER'.

    For each field tag it adds an entry to the returned dict with the tag as the key and a list of the entries as the value, the list has each line separately, so for the following string in a record:

    "AF BREVIK, I

        ANICIN, B"

    The entry in the returned dict would be `{'AF' : ["BREVIK, I", "ANICIN, B"]}`

    [Record](#metaknowledge.Record) objects can be created with these dictionaries as the initializer.

    # Parameters

    _paper_ : `file stream`

    > An open file, with the current line at the beginning of the record.

    # Returns

    `dict[str : List[str]]`

    > A dictionary mapping WOS tags to lists, the lists are of strings, each string is a line of the record associated with the tag.
    """
    tagList = []
    doneReading = False
    for l in paper:
        if len(l[1]) < 3:
            #Line too short
            raise BadISIRecord("Missing field on line " + str(l[0]) + " : " + l[1])
        elif 'ER' in l[1][:2]:
            #Reached the end of the record
            doneReading = True
            break
        elif l[1][2] != ' ':
            #Field tag longer than 2 or offset in some way
            raise BadISIFile("Field tag not formed correctly on line " + str(l[0]) + " : " + l[1])
        elif '   ' in l[1][:3]: #the string is three spaces in row
            #No new tag append line to current tag (last tag in tagList)
            tagList[-1][1].append(l[1][3:-1])
        else:
            #New tag create new entry at the end of tagList
            tagList.append((l[1][:2], [l[1][3:-1]]))
    if not doneReading:
        raise BadISIRecord("End of file reached before ER: " + l[1])
    else:
        retdict = collections.OrderedDict(tagList)
        if len(retdict) == len(tagList):
            return retdict
        else:
            dupSet = set()
            for tupl in tagList:
                if tupl[0] in retdict:
                    dupSet.add(tupl[0])
            raise BadISIRecord("Duplicate tags (" + ', '.join(dupSet) + ") in record")



"""
"""
from .constants import tagsAndNames
from .graphHelpers import _ProgressBar
from .recordCollection import RecordCollection

import networkx as nx
import metaknowledge
"""
TODO: FIGURE OUT
def diffusion(source, target):
    targetCounts = diffusionCount(source, target)
    sourceCounts = diffusionCount(source, source)
"""

def diffusionGraph(source, target, sourceType = "raw", targetType = "raw"):
    """Takes in two [`RecordCollections`](#RecordCollection.RecordCollection) and produces a graph of the citations of the `Records` of _source_ by the `Records` of _target_. By default the graph is of `Record` objects but this can be changed with the _sourceType_ and _targetType_ keywords.

    Each node on the graph has two boolean attributes, `"source"` and `"target"` indicating if they are targets or sources. Note, if the types of the sources and targets are different the attributes will not be checked for overlap of the other type. e.g. if the source type is `'TI'` (title) and the target type is `'UT'` (WOS number), and there is some overlap of the targets and sources. Then the Record corresponding to a source node will not be checked for being one of the titles of the targets, only its WOS number will be considered.

    # Parameters

    _source_ : `RecordCollection`

    >A metaknowledge `RecordCollection` containing the `Records` being cited

    _target_ : `RecordCollection`

    >A metaknowledge `RecordCollection` containing the `Records` citing those in _source_

    _sourceType_ : `str`

    >default `'raw'`, if `'raw'` the returned graph will contain `Records` as source nodes. If it is a WOS tag of the long name of one then the nodes will be of that type.

    _targetType_ : `str`

    >default `'raw'`, if `'raw'` the returned graph will contain `Records` as target nodes. If it is a WOS tag of the long name of one then the nodes will be of that type.

    # Returns

    `networkx Directed Graph`

    >A directed graph of the diffusion network
    """
    if sourceType != "raw" and sourceType not in tagsAndNames:
        raise RuntimeError("{} is not a valid node type, only 'raw' or those strings in tagsAndNames are allowed".format(nodeType))
    if targetType != "raw" and targetType not in tagsAndNames:
        raise RuntimeError("{} is not a valid node type, only 'raw' or those strings in tagsAndNames are allowed".format(nodeType))
    if metaknowledge.VERBOSE_MODE:
        PBar = _ProgressBar(0, "Starting to make a diffusion network")
        count = 0
        maxCount = len(source)
    else:
        PBar = None
    sourceDict = {}
    workingGraph = nx.DiGraph()
    for Rs in source:
        if PBar:
            count += 1
            PBar.updateVal(count / maxCount * .25, "Analyzing source: " + str(Rs))
        RsVal, RsExtras = makeNodeID(Rs, sourceType)
        if RsVal:
            sourceDict[Rs.createCitation()] = RsVal
            for val in RsVal:
                if val not in workingGraph:
                    workingGraph.add_node(val, source = True, target = False, **RsExtras)
    if PBar:
        count = 0
        maxCount = len(target)
        PBar.updateVal(.25, "Done analyzing sources, starting on targets")
    for Rt in target:
        RtVal, RtExtras = makeNodeID(Rt, targetType)
        if PBar:
            count += 1
            PBar.updateVal(count / maxCount * .75 + .25, "Analyzing target: " + str(Rt))
        if RtVal:
            for val in RtVal:
                if val not in workingGraph:
                    workingGraph.add_node(val, source = False, target = True, **RtExtras)
                else:
                    workingGraph.node[val]["target"] = True
                targetCites = Rt.CR
                if targetCites:
                    for Rs in (sourceDict[c] for c in targetCites if c in sourceDict):
                        for sVal in Rs:
                            workingGraph.add_edge(val, sVal)
    if PBar:
        PBar.finish("Done making a diffusion network of {} sources and {} targets".format(len(source), len(target)))
    return workingGraph


def diffusionCount(source, target, sourceType = "raw", pandasFriendly = False,  compareCounts = False, _ProgBar = None):
    """Takes in two [`RecordCollections`](#RecordCollection.RecordCollection) and produces a `dict` counting the citations of the `Records` of _source_ by the `Records` of _target_. By default the `dict` uses `Record` objects as keys but this can be changed with the _sourceType_ keyword to any of the WOS tags.

    # Parameters

    _source_ : `RecordCollection`

    >A metaknowledge `RecordCollection` containing the `Records` being cited

    _target_ : `RecordCollection`

    >A metaknowledge `RecordCollection` containing the `Records` citing those in _source_

    _sourceType_ : `optional [str]`

    >default `'raw'`, if `'raw'` the returned `dict` will contain `Records` as keys. If it is a WOS tag of the long name of one then the keys will be of that type.

    _pandasFriendly_ : `optional [bool]`

    > default `False`, makes the output be a dict with two keys one `"Record"` is the list of Records ( or data type requested by _sourceType_) the other is their occurence counts as `"Counts"`.

    _compareCounts_ : `optional [boo]`

    > default `False`, if `True` the diffusion analysis will be run twice, first with source and target setup like the default (global scope) then using only the source `RecordCollection` (local scope).

    # Returns

    `dict[:int]`

    > A dictionary with the type given by _sourceType_ as keys and integers as values, by default. If _compareCounts_ is `True` the values are tuples with the first integer being the diffusion in the target and the second the diffusion in the source.

    > If _pandasFriendly_ is `True` the returned dict has keys with the names of the WOS tags and lists with their values, i.e. a table with labled columns. The counts are in the column named `"Count"` and if _compareCounts_ the local count is in a column called `"LocalCount"`.
    """
    if sourceType != "raw" and sourceType not in tagsAndNames:
        raise RuntimeError("{} is not a valid node type, only 'raw' or those strings in tagsAndNames are allowed".format(nodeType))
    if not isinstance(source, RecordCollection) or not isinstance(target, RecordCollection):
        raise RuntimeError("Source and target must be RecordCollections.")
    if metaknowledge.VERBOSE_MODE or _ProgBar:
        if _ProgBar:
            PBar = _ProgBar
            PBar.updateVal(0, "Starting to analyse a diffusion network")
        else:
            PBar = _ProgressBar(0, "Starting to analyse a diffusion network")
        count = 0
        maxCount = len(source)
    else:
        PBar = None
    if PBar:
        count = 0
        maxCount = len(source)
        PBar.updateVal(.25, "Done analyzing sources, starting on targets")
    sourceDict = {}
    sourceSet = set()
    for Rs in source:
        if PBar:
            count += 1
            PBar.updateVal(count / maxCount * .25, "Analyzing source: " + str(Rs))
        RsVal, RsExtras = makeNodeID(Rs, sourceType)
        if RsVal:
            sourceDict[Rs.createCitation()] = RsVal
            sourceSet.update(RsVal)
    sourceCounts = {s : 0 for s in sourceSet}
    if PBar:
        count = 0
        maxCount = len(target)
        PBar.updateVal(.25, "Done analyzing sources, starting on targets")
    for Rt in target:
        if PBar:
            count += 1
            PBar.updateVal(count / maxCount * .75 + .25, "Analyzing target: " + str(Rt))
        targetCites = Rt.CR
        if targetCites:
            for Rs in (sourceDict[c] for c in targetCites if c in sourceDict):
                for sVal in Rs:
                    sourceCounts[sVal] += 1
    if compareCounts:
        localCounts = diffusionCount(source, target, sourceType = sourceType, pandasFriendly = False,  compareCounts = False, _ProgBar = PBar)
    if PBar and not _ProgBar:
        PBar.finish("Done counting the diffusion of {} sources into {} targets".format(len(source), len(target)))
    if pandasFriendly:
        retDict = {"Count" : []}
        if compareCounts:
            retDict["LocalCount"] = []
        if sourceType == 'raw':
            retrievedFields = []
            for R in sourceCounts.keys():
                tagsLst = [t for t in R.activeTags() if t not in retrievedFields]
                retrievedFields += tagsLst
            for tag in retrievedFields:
                retDict[tag] = []
            for R, occ in sourceCounts.items():
                Rvals = R.getTagsDict(retrievedFields, cleaned = True)
                for tag in retrievedFields:
                    retDict[tag].append(Rvals[tag])
                retDict["Count"].append(sourceCounts[R])
                if compareCounts:
                    retDict["LocalCount"].append(localCounts[R])
        else:
            countLst = []
            recLst = []
            locLst = []
            for R, occ in sourceCounts.items():
                countLst.append(occ)
                recLst.append(R)
                if compareCounts:
                    locLst.append(localCounts[R])
            if compareCounts:
                retDict = {sourceType : recLst, "Count" : countLst, "localCount" : locLst}
            else:
                retDict = {sourceType : recLst, "Count" : countLst}
        return retDict
    else:
        if compareCounts:
            for R, occ in localCounts.items():
                sourceCounts[R] = (sourceCounts[R], occ)
        return sourceCounts

def makeNodeID(Rec, ndType, extras = None):
    """Helper to make a node ID, extras is currently not used"""
    if ndType == 'raw':
        recID = Rec
    else:
        recID =  getattr(Rec, ndType)
    if recID is None:
        pass
    elif isinstance(recID, list):
        recID = tuple(recID)
    else:
        recID = (recID, )
    extraDict = {}
    if extras:
        for tag in extras:
            if tag == "raw":
                extraDict[Tag] = Rec
            else:
                extraDict[Tag] = getattr(Rec, tag)
    return recID, extraDict



from .journalAbbreviations import getj9dict, manaulDBname, addToDB

abbrevDict = None

class BadCitation(Warning):
    """
    Exception thrown by Citation
    """
    pass

class Citation(object):
    """
    A class to hold citation strings and allow for comparison between them.

    The initializer takes in a string representing a WOS citation they are in the form:

    > Author, Year, Journal, Volume, Page, DOI

    Author is the author's name in the form of first last name first initial sometimes followed by a period.
    Year is the year of publication.
    Journal being the 29-Character Source Abbreviation of the journal.
    Volume is the volume number(s) of the publication preceded by a V
    Page is the page number the record starts on
    DOI is the DOI number of the cited record preceeded by the letters "DOI"
    Combined they look like:

    > Nunez R., 1998, MATH COGNITION, V4, P85, DOI 10.1080/135467998387343

    Note that any of the fields have been known to be missing and the requirements for the fields are not always met. If something is in the source string that cannot be interpeted as any of these it is put in the `misc` attribute.

    The reason for this class is that the WOS data are often irregular. It is designed to allow comparison between WOS citation strings, even when they are missing pieces.

    # Customizations

    Citation's hashing and equality checking are based on [`getID()`](#Citation.getID) and use the values of `author`, `year` and `journal`.

    When converted to a string a Citation will return the original string.

    # Attributes

    As noted above, citations are considered to be divided into six distinct fields (Author, Year, Journal, Volume, Page and DOI) with a seventh misc for anything not in those. Records thus have an attribute with a name corresponding to each `author`, `year`, `journal`, `V`, `P`, `DOI` and `misc` respectively. These are created if there is anything in the field. So a Citation created from the string: "Nunez R., 1998, MATH COGNITION" would have `author`, `year` and `journal` defined. While one from "Nunez R." would have only the attribute `misc`.

    If the parsing of a citation string fails the attribute `bad` is set to True and the attribute `error` is created to contain the error, which is a [BadCitation](#metaknowledge.BadCitation) object. If no errors occur `bad` is `False`.

    The attribute `original` is the unmodified string (_cite_) given to create the Citation, it can also be accessed by converting to a string, e.g. with `str()`.

    # \_\_Init\_\_

    Citations can be created by [Records](#metaknowledge.Record) or by giving the initializer a string containing a WOS style citation.

    # Parameters

    _cite_ : `str`

    > A str containing a WOS style citation.
    """
    def __init__(self, cite):
        #save original
        self.original = cite
        #setup attributes
        self.bad = False
        self._isjourn = None
        self._hash = None
        self.author = None
        self.year = None
        self.P = None
        self.V = None
        self.journal = None
        self.DOI = None
        self.misc = []
        #split by citation seperator
        c = ' '.join(cite.upper().split()).split(', ')
        if 'DOI' in c[-1][:3]:
            self.DOI = c.pop().split(' ')[-1]
        else:
            self.DOI = None
        if len(c) < 2:
            self.bad = True
            self.error = BadCitation("Too few elements")
        while len(c) > 0:
            field = c.pop(0).upper()
            if field.isnumeric():
                self.year = int(field)
            elif not self.author:
                self.author = field.replace('.','').title()
            elif not self.journal:
                self.journal = field
            elif field[0] == 'V' and field[1:].isnumeric():
                self.V = field
            elif 'P' == field[0] and field[1:].isnumeric():
                self.P = field
            else:
                self.misc.append(field)
        if not self.author or not self.year or not self.journal and not self.bad:
            self.bad = True
            self.error = BadCitation("Not a complete set of author, year and journal")

    def __str__(self):
        """
        returns the original string
        """
        return self.original

    def __hash__(self):
        """
        A hash for Citation that should be equal to the hash of other citations that are equal to it. Based on the values returned by [`getID()`](#Citation.getID).
        """
        if self._hash:
            return self._hash
        else:
            self._hash = hash(self.getID())
            return self._hash

    def __eq__(self, other):
        """
        First checks DOI for equality then checks each attribute if any are not equal False is returned
        """
        if not isinstance(other, Citation):
            return NotImplemented
        return hash(self) == hash(other)

    def __ne__(self, other):
        """
        Returns the inverse of equality
        """
        return not self == other

    def __repr__(self):
        """
        the representation of the Citation is its original form
        """
        return self.original

    def isAnonymous(self):
        """
        Checks if the author is given as "[ANONYMOUS]" and returns `True` if so.

        # Returns

        `bool`

        > True if the author is ANONYMOUS otherwise `False`.
        """
        return self.author == "[Anonymous]"

    def getID(self):
        """
        Returns all of "author, year, journal" available. It is for shortening labels when creating networks as the resultant strings are often unique. [`getExtra()`](#Citation.getExtra) gets everything not returned by `getID()`.

        This is also used for hashing and equality checking.

        # Returns

        `str`

        > A string to use as the shortened ID of a node.
        """
        if self.bad:
            atrLst = []
            if self.author:
                atrLst.append(self.author)
            if self.year:
                atrLst.append(str(self.year))
            if self.journal:
                atrLst.append(self.journal)
            return ', '.join(atrLst)
        else:
            return "{0}, {1}, {2}".format(self.author, self.year, self.journal)

    def getExtra(self):
        """
        Returns any V, P, DOI or misc values as a string. These are all the values not returned by [`getID()`](#Citation.getID).

        # Returns

        `str`

        > A string containing the data not in the ID of the Citation.
         """
        extraTags = ['V', 'P', 'DOI', 'misc']
        retVal = ""
        for tag in extraTags:
            if getattr(self, tag):
                retVal += getattr(self, tag) + ', '
        if len(retVal) > 2:
            return retVal[:-2]
        else:
            return retVal

    def isJournal(self, manaulDB = manaulDBname, returnDict = 'both', checkIfExcluded = False):
        """Returns `True` if the Citation's journal field is a journal abbreviation given by WOS, i.e. checks if the citation is citing a journal. Requires the j9Abbreviations database file.

        # Returns

        `bool`

        > `True` if the Citation is for a journal
        """
        global abbrevDict
        if abbrevDict is None:
            abbrevDict = getj9dict(manualDB = manaulDB, returnDict = returnDict)
        if checkIfExcluded and self.journal:
            try:
                if abbrevDict.get(self.journal, [True])[0]:
                    return False
                else:
                    return True
            except IndexError:
                return False
        elif self._isjourn is None:
            if self.journal:
                dictVal = abbrevDict.get(self.journal, [b''])[0]
                if dictVal:
                    self._isjourn = dictVal
                else:
                    self._isjourn = False
            else:
                self._isjourn = False
        return self._isjourn

    def getFullJournalName(self):
        """Returns the full name of the Citation's journal field. Requires the j9Abbreviations database file.

        # Returns

        `str`

        > The first full name given for the journal of the Citation (or the first name in the WOS list if multiple names exist), if there is not one then `None` is returned
        """
        global abbrevDict
        if abbrevDict is None:
            abbrevDict = getj9dict()
        if self.isJournal():
            return abbrevDict[self.journal][0]
        else:
            return None

    def addToDB(self, manualName = None, manaulDB = manaulDBname, invert = False):
        """Adds the journal of this Citation to the user created database of journals. This will cause [isJournal()](#Citation.isJournal) to return `True` for this Citation and all others with its `.journal`.

        # Parameters

        _manualName_ : `optional [str]`

        > Default `None`, the full name of journal to use. If not provided the full name will be the same as the abbreviation.

        _manaulDB_ : `optional [str]`

        > The name of the database file, the default is [metaknowledge.journalAbbreviations.manaulDBname](#journalAbbreviations.manaulDBname)

        _invert_ : `optional [bool]`

        > Default `False`, if `True` the journal will be removed instead of added
        """
        try:
            if invert:
                d = {self.journal : ''}
            elif manualName is None:
                d = {self.journal : self.journal}
            else:
                d = {self.journal : manualName}
            addToDB(abbr = d, dbname = manaulDB)
        except KeyError:
            raise KeyError("This citation does not have a journal field.")
        else:
            global abbrevDict
            abbrevDict.update(d)

def filterNonJournals(citesLst, invert = False):
    """Removes the Citations from _citesLst_ that are not journals.

    # Parameters

    _citesLst_ : `list [Citation]`

    > A list of citations to be filtered

    _invert_ : `optional [bool]`

    > Default `False`, if `True` non-journals will be kept istead of journals

    # Returns

    `list [Citation]`

    > A filtered list of Citations from _citesLst_
    """

    retCites = []
    for c in citesLst:
        if c.isJournal():
            if not invert:
                retCites.append(c)
        elif invert:
            retCites.append(c)
    return retCites
